import asyncio
import base64
import collections
import json
import logging
import os
import shutil
import smtplib
import types
import warnings

from datetime import datetime
from collections import OrderedDict
from fnmatch import fnmatch
from pathlib import Path

from email.mime.text import MIMEText


from urllib import parse

from concurrent.futures import ThreadPoolExecutor

from pypeman.message import Message
from pypeman.channels import Dropped
from pypeman.channels import Rejected
from pypeman.persistence import get_backend

logger = logging.getLogger(__name__)

# All declared nodes registered here
all_nodes = []
node_by_name = {}

# Can be redefined
default_thread_pool = ThreadPoolExecutor(max_workers=3)

SENTINEL = object()


def choose_first_not_none(*args):
    """ Choose first non None alternative in args.
    :param args: alternative list
    :return: the first non None alternative.
    """
    for a in args:
        if a is not None:
            return a
    return None


def callable_or_value(val, msg):
    """
    Return `val(msg)` if value is a callable else `val`.
    """
    if callable(val):
        name = val(msg)
    else:
        name = val
    return name


def get_context(msg, date=None, counter=None):
    cdate = date or datetime.now()
    timestamp = msg.timestamp
    context = {'year': cdate.year,
               'month': cdate.month,
               'day': cdate.day,
               'hour': cdate.hour,
               'second': cdate.second,
               'msg_year': timestamp.year,
               'msg_month': timestamp.month,
               'msg_day': timestamp.day,
               'msg_hour': timestamp.hour,
               'msg_second': timestamp.second,
               'msg_uid': msg.uuid,
               'counter': counter
               }
    return context


class NodeException(Exception):
    """ custom exception """


def find_node(name="", match=None):
    """
    find nodes in a pypeman graph

    :param name:  name of node to search
                  name can contain * ? as specified in fnmatch.fnmatch
    """

    if name:
        return next(
            filter(lambda node: fnmatch(node.name, name), all_nodes),
            None,
            )
    return next(filter(match, all_nodes, None))


class BaseNode:
    """
    Base of all Nodes.
    If you create a new node, you must inherit from this class and
    implement `process` method.

    :func:`save_data <pypeman.nodes.BaseNode.save_data>`

    :param name: Name of node. Used in log or test.
    :param log_output: To enable output logging for this node.
    :store_output_as: Store output message in msg.ctx as specified key
    :store_input_as: Store input message in msg.ctx as specified key
    :passthrough: If True, node is executed but output message is same as input

    """

    _used_names = set()  # already used node names to ensure uniqueness

    def __init__(self, *args, name=None, log_output=False, **kwargs):
        cls = self.__class__
        self.channel = None

        all_nodes.append(self)

        name = name or cls.__name__ + "_" + str(len(all_nodes))
        if name in cls._used_names:
            raise NodeException(
                "can't create Node with name %s. It exists already" % name)
        cls._used_names.add(name)
        node_by_name[name] = self
        self.name = name

        self.store_output_as = kwargs.pop('store_output_as', None)
        self.store_input_as = kwargs.pop('store_input_as', None)
        self.passthrough = kwargs.pop('passthrough', None)
        self.next_node = None

        self.processed = 0

        if log_output:
            # Enable logging
            self._handle_without_log = self.handle
            setattr(self, 'handle', self._log_handle)

    def fullpath(self):
        """
        Return the channel name and node name dot concatened.
        """
        return "%s.%s" % (self.channel.name, self.name)

    async def handle(self, msg):
        """ Handle message is called by channel to launch process method on it.
        Some other structural processing take place here.
        Please, don't modify unless you know what you are doing.

        :param msg: incoming message
        :return: modified message after a process call and some treatment
        """

        # TODO : Make sure exceptions are well raised (does not happen if i.e 1/0 here atm)
        if self.store_input_as:
            msg.add_context(self.store_input_as, msg)
        if self.passthrough:
            old_msg = msg.copy()
        # Allow process as coroutine function
        if asyncio.iscoroutinefunction(self.process):
            result = await self.async_run(msg)
        else:
            result = self.run(msg)

        self.processed += 1

        if isinstance(result, asyncio.Future):
            result = await result

        self.channel.logger.debug(
            '%s node end handle msg %s, result is msg %s',
            str(self), str(msg), str(result))

        if self.next_node:
            if isinstance(result, types.GeneratorType):
                gene = result
                result = msg  # Necessary if all nodes result are dropped
                for res in gene:
                    try:
                        result = await self.next_node.handle(res)
                    except Dropped:
                        pass
                    # TODO Here result is last value returned. Is it a good idea ?
            else:
                if self.store_output_as:
                    result.add_context(self.store_output_as, result)

                if self.passthrough:
                    result.payload = old_msg.payload
                    result.meta = old_msg.meta

                result = await self.next_node.handle(result)

        return result

    async def _log_handle(self, msg):
        """
        Used when node logging is enabled. Log after node processing.
        """
        result = await self._handle_without_log(msg)

        # Log message
        result.log(logger=self.channel.logger, log_level=logging.DEBUG)

        return result

    async def _test_handle(self, msg):
        """ Specific handle for TEST mode to enable some testing and introspection operations
        like mock input and/or output, or count processed message.

        :param msg: Message to process.
        :return: Processed message.
        """
        # Keep last input
        self._last_input = msg.copy()

        if self._mock_input:
            if callable(self._mock_input):
                msg = self._mock_input(msg)
            else:
                msg = self._mock_input

        result = await self._handle(msg)

        return result

    async def async_run(self, msg):
        """ Used to overload behaviour like thread Node without rewriting handle process """
        result = await self.process(msg)
        return result

    def run(self, msg):
        """ Used to overload behaviour like thread Node without rewriting handle process """
        result = self.process(msg)
        return result

    def process(self, msg):
        """ Implement this function in child classes to create
        a new Node.

        :param msg: The incoming message
        :return: The processed message
        """
        return msg

    async def save_data(self, key, value):
        """
        Save data in configured persistence backend for next usage.

        :param key: Key of saved data.
        :param value: Value saved.
        """
        await (await get_backend(self.channel.loop)).store(self.fullpath(), key, value)

    async def restore_data(self, key, default=SENTINEL):
        """
        Restore previously saved data from configured persistence backend.

        :param key: Key of restored data.
        :param default: if key is missing, don't raise exception and return this value instead.
        :return: Saved data if exist or default value if specified.
        """
        if default is not SENTINEL:
            return await (await get_backend(self.channel.loop)).get(self.fullpath(), key, default)
        else:
            return await (await get_backend(self.channel.loop)).get(self.fullpath(), key)

    # Allow to mock input or
    def mock(self, input=None, output=None):
        """
        Allow to mock input or output of a node for testing purpose.

        :param input: A message to replace the input in this node.
        :param output: A return message to replace processing of this mock.
        """
        if input:
            self._mock_input = input

        if output:
            if not hasattr(self, '_orig_process'):
                self._orig_process = self.process

            def new_process(msg):
                if callable(output):
                    return output(msg)
                else:
                    return output

            self.process = new_process

    def _reset_test(self):
        """ Set test mode and reset test information """
        self.processed = 0

        if not hasattr(self, '_handle'):
            self._handle = self.handle
            setattr(self, 'handle', self._test_handle)

        if hasattr(self, '_orig_process'):
            self.process = self._orig_process

        self._mock_input = None
        self._last_input = None

    def last_input(self):
        return self._last_input

    def __str__(self):
        return "<%s(%s)>" % (self.channel.name, self.name)


class RaiseError(BaseNode):
    def process(self, msg):
        raise Exception("Test node")


class Drop(BaseNode):
    """ Use this node to tell the channel the message is Dropped. """
    def __init__(self, message=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.message = message

    def process(self, msg):
        if self.message:
            raise Dropped(self.message)
        else:
            raise Dropped()


class DropNode(Drop):
    def __init__(self, *args, **kwargs):
        warnings.warn("DropNode node is deprecated. Replace it by Drop node", DeprecationWarning)
        super().__init__(*args, **kwargs)


class Empty(BaseNode):
    """ Return an empty new message. """
    def process(self, msg):
        return Message()


class SetCtx(BaseNode):
    """ Push the message in the context with the key `ctx_name` """

    def __init__(self, ctx_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ctx_name = ctx_name

    def process(self, msg):
        msg.meta = msg.ctx[self.ctx_name]['meta']
        msg.payload = msg.ctx[self.ctx_name]['payload']

        return msg


class ThreadNode(BaseNode):
    """
    Inherit from this class instead of BaseNode to avoid
    long run node blocking main event loop.
    """

    def __init__(self, *args, thread_pool=None, **kwargs):
        super().__init__(*args, **kwargs)

        if thread_pool is None:
            self.executor = default_thread_pool
        else:
            self.executor = thread_pool

    def run(self, msg):
        result = self.channel.loop.run_in_executor(self.executor, self.process, msg)

        return result


class Log(BaseNode):
    """
    Node to show some information about node, channel and message. Use for debug.
    """
    def __init__(self, *args, **kwargs):
        """
            Node specific params
            :param level: log level of node
            :param show_ctx: whether to log the context
        """
        self.lvl = kwargs.pop('level', logging.INFO)
        self.show_ctx = kwargs.pop('show_ctx', False)

        super().__init__(*args, **kwargs)

    def process(self, msg):
        self.channel.logger.log(self.lvl, '%s %s', str(self), str(msg))

        if self.channel.parent_uids:
            self.channel.logger.log(self.lvl, 'Parent channels: %s', ', '.join(self.channel.parent_names))

        msg.log(logger=self.channel.logger, payload=True, meta=True,
                context=self.show_ctx, log_level=self.lvl)

        return msg


class Sleep(BaseNode):
    """ Wait `duration` seconds before returning message."""
    def __init__(self, *args, duration=1, **kwargs):
        self.duration = duration
        super().__init__(*args, **kwargs)

    async def process(self, msg):
        await asyncio.sleep(self.duration)
        return msg


class JsonToPython(BaseNode):
    """ Convert json message payload to python dict."""
    # TODO encoding management
    def __init__(self, *args, encoding='utf-8', **kwargs):
        self.encoding = encoding
        super().__init__(*args, **kwargs)

    def process(self, msg):
        encoded_payload = msg.payload.encode(self.encoding)
        msg.payload = json.loads(encoded_payload)
        msg.content_type = 'application/python'
        return msg


class PythonToJson(BaseNode):
    """ Convert python payload to json."""
    def __init__(self, *args, encoding='utf-8', indent=None, **kwargs):
        self.encoding = encoding
        super().__init__(*args, **kwargs)
        self.indent = indent

    def process(self, msg):
        msg.payload = json.dumps(msg.payload, indent=self.indent)
        msg.content_type = 'application/json'
        return msg


class Encode(BaseNode):
    """ Encode payload in specified encoding to byte.
    """
    def __init__(self, *args, **kwargs):
        self.encoding = kwargs.pop('encoding', 'utf-8')
        super().__init__(*args, **kwargs)

    def process(self, msg):
        msg.payload = msg.payload.encode(self.encoding)
        return msg


class Decode(BaseNode):
    """ Decode payload from byte to specified encoding
    """
    def __init__(self, *args, **kwargs):
        self.encoding = kwargs.pop('encoding', 'utf-8')
        super().__init__(*args, **kwargs)

    def process(self, msg):
        msg.payload = msg.payload.decode(self.encoding)
        return msg


class B64Encode(BaseNode):
    """ Encode payload in specified encoding to byte.
    """
    def __init__(self, *args, altchars=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.altchars = altchars

    def process(self, msg):
        msg.payload = base64.b64encode(msg.payload, altchars=self.altchars)
        return msg


class B64Decode(BaseNode):
    """ Decode payload from byte to specified encoding
    """
    def __init__(self, *args, altchars=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.altchars = altchars

    def process(self, msg):
        msg.payload = base64.b64decode(msg.payload, altchars=self.altchars)
        return msg


# TODO put Save in specific file ?
class SaveNullBackend():
    """ For testing purpose """
    def store(self, message):
        pass


class SaveFileBackend():
    """ Backend used to store message with ``Save`` node.
    """
    def __init__(self, path, filename, channel):
        self.path = path
        self.filename = filename
        self.counter = 0
        self.channel = channel

    def store(self, message):
        today = datetime.now()
        timestamp = message.timestamp

        context = {'counter': self.counter,
                   'year': today.year,
                   'month': today.month,
                   'day': today.day,
                   'hour': today.hour,
                   'second': today.second,
                   'msg_year': timestamp.year,
                   'msg_month': timestamp.month,
                   'msg_day': timestamp.day,
                   'msg_hour': timestamp.hour,
                   'msg_second': timestamp.second,
                   'muid': message.uuid,
                   'cuid': getattr(self.channel, 'uuid', '???')
                   }

        filepath = os.path.join(self.path, self.filename % context)

        try:
            # Make missing dir if any
            os.makedirs(os.path.dirname(filepath))
        except FileExistsError:
            pass

        with open(filepath, 'w') as file_:
            file_.write(message.payload)

        self.counter += 1


class Save(ThreadNode):
    """ Save a message in specified uri """
    def __init__(self, *args, uri=None, **kwargs):

        self.uri = uri
        parsed = parse.urlparse(self.uri)

        super().__init__(*args, **kwargs)

        if parsed.scheme == 'file':
            filename = parsed.query.split('=')[1]

            self.backend = SaveFileBackend(path=parsed.path, filename=filename, channel=self.channel)
        else:
            self.backend = SaveNullBackend()

    def process(self, msg):
        self.backend.store(msg)
        return msg


class MessageStore(Save):
    def __init__(self, *args, **kwargs):
        warnings.warn("MessageStore node is deprecated. Replace it by Save node", DeprecationWarning)
        super().__init__(*args, **kwargs)


class FileReader(BaseNode):
    """ Reads a file and sets payload to the file's contents. """
    def __init__(self, filename=None, filepath=None, binary_file=False, *args, **kwargs):
        self.filename = filename
        self.filepath = filepath
        self.binary_file = binary_file
        self.counter = 0
        if self.filename:
            warnings.warn("Filename deprecated, use filepath instead", DeprecationWarning)
        super().__init__(*args, **kwargs)

    def process(self, msg):
        self.counter += 1
        if self.filepath:
            filepath = callable_or_value(self.filepath, msg)

        elif self.filename:
            filename = callable_or_value(self.filename, msg)

            path = os.path.dirname(msg.meta['filepath'])
            filepath = os.path.join(path, filename)
        else:
            filepath = msg.meta['filepath']

        context = get_context(msg=msg, counter=self.counter)
        filepath = filepath % context
        name = os.path.basename(filepath)

        if self.binary_file:
            mode = "rb"
        else:
            mode = "r"

        with open(filepath, mode) as file:
            msg.payload = file.read()
            msg.meta['filename'] = name
            msg.meta['filepath'] = filepath

        return msg


class FileWriter(BaseNode):
    """
        Write a file with the message content.
        Can create a validation file with no content but with same path and same
        base name with different extension (for example .ok)
    """
    def __init__(self, filepath=None, binary_mode=False, safe_file=True, create_valid_file=False,
                 validation_extension=".ok", *args, **kwargs):
        self.filepath = filepath
        self.binary_mode = binary_mode
        self.safe_file = safe_file
        self.create_valid_file = create_valid_file
        self.validation_extension = validation_extension
        self.first_filename = True
        self.counter = 0
        super().__init__(*args, **kwargs)

    def process(self, msg):
        self.counter += 1
        meta_filepath = msg.meta.get('filepath')

        if self.filepath:
            filepath = callable_or_value(self.filepath, msg)
        else:
            filepath = meta_filepath

        if not filepath:
            raise ValueError("filepath must be defined in parameters or in msg.meta")

        context = get_context(msg=msg, counter=self.counter)
        dest = filepath % context
        old_file = dest
        if self.safe_file:
            dest = old_file + '.tmp'
        with open(dest, 'w' + ('b' if self.binary_mode else '')) as file_:
            file_.write(msg.payload)
        if self.safe_file:
            os.rename(dest, old_file)
        if self.create_valid_file:
            validation_path = Path(old_file).with_suffix(self.validation_extension)
            validation_path.touch()
        return msg


class FileMover(BaseNode):
    """
    Move file

    Used to store files at another place
    params:
    dest_fpath : path of the destination folder (if it doesn't exists it creates it)
    """
    def __init__(self, dest_path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dest_path = dest_path
        if not os.path.exists(dest_path):
            os.makedirs(dest_path)

    def process(self, msg):
        dest_fpath = os.path.join(self.dest_path, msg.meta["filename"])
        logger.info("move file %s to dest %s", msg.meta["filepath"], dest_fpath)
        shutil.move(msg.meta["filepath"], dest_fpath)
        msg.meta["filepath"] = dest_fpath
        return msg


class FileCleaner(BaseNode):
    """
    Delete a file and/or all metafiles in a directory with same basename but
    with a given extension

    param:
    extensions_to_rm => list of all extensions to rm (example: [".ok"])
    msg.meta["filepath"]
    """
    def __init__(self, extensions_to_rm=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.extensions_to_rm = extensions_to_rm

    def process(self, msg):
        fpath = Path(msg.meta['filepath'])
        if fpath.is_file():
            logger.info("delete %s ...", msg.meta["filepath"])
            fpath.unlink()
        if self.extensions_to_rm:
            for extension in self.extensions_to_rm:
                meta_path = fpath.with_suffix(extension)
                if meta_path.is_file():
                    logger.info("delete %s ...", (meta_path))
                    meta_path.unlink()


class Map(BaseNode):
    """ Used to map input message keys->values to another keys->values """

    def __init__(self, *args, **kwargs):
        self.mapping = kwargs.pop('mapping')
        self.recopy = kwargs.pop('recopy')
        path = kwargs.pop('path', "")

        self.path = 'payload'

        if path:
            self.path += '.' + path

        super().__init__(*args, **kwargs)

    def process(self, msg):
        current = msg
        parts = self.path.split('.')
        for part in parts:
            try:
                current = current[part]
            except (TypeError, KeyError):
                current = getattr(current, part)

        old_dict = current
        new_dict = {}

        for mapItem in self.mapping:
            mapItem.conv(old_dict, new_dict, msg)
        if self.recopy:
            new_dict.update(old_dict)

        dest = msg
        for part in parts[:-1]:
            try:
                dest = dest[part]
            except (TypeError, KeyError):
                dest = getattr(dest, part)

        try:
            dest[parts[-1]] = new_dict
        except KeyError:
            setattr(dest, parts[-1], new_dict)

        return msg


class MappingNode(Map):
    def __init__(self, *args, **kwargs):
        warnings.warn("MappingNode node is deprecated. Replace it by 'Map' node", DeprecationWarning)
        super().__init__(*args, **kwargs)


class ToOrderedDict(BaseNode):
    """ this node yields an ordered dict with the keys 'keys' and the values from the payload
       if the payload does not contain certain values defaults can be specified with defaults
    """
    NONE = object()

    def __init__(self, *args, **kwargs):
        self.keys = kwargs.pop('keys')
        defaults = kwargs.pop('defaults', dict())
        self.dflt_dict = OrderedDict()
        for key in self.keys:
            self.dflt_dict[key] = defaults.get(key, ToOrderedDict.NONE)
        path = kwargs.pop('path', None)
        self.path = 'payload'
        if path:
            self.path += '.' + path
        super().__init__(*args, **kwargs)

    def process(self, msg):
        current = msg
        parts = self.path.split('.')

        for part in parts:
            try:
                current = current[part]
            except (TypeError, KeyError):
                current = getattr(current, part)
        old_dict = current
        new_dict = OrderedDict()

        for key in self.keys:
            val = old_dict.get(key, ToOrderedDict.NONE)
            if val is ToOrderedDict.NONE:
                val = self.dflt_dict[key]
            if val is not ToOrderedDict.NONE:
                new_dict[key] = val

        dest = msg
        for part in parts[:-1]:
            try:
                dest = dest[part]
            except (TypeError, KeyError):
                dest = getattr(dest, part)

        try:
            dest[parts[-1]] = new_dict
        except KeyError:
            setattr(dest, parts[-1], new_dict)

        return msg


class Email(ThreadNode):
    """ Node that send Email.
    """
    def __init__(self, *args, host=None, port=None, user=None, password=None, ssl=False, start_tls=False,
                 subject=None, sender=None, recipients=None, content=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.subject = subject or ""
        self.sender = sender
        self.recipients = recipients
        self.content = content
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.start_tls = start_tls
        self.ssl = ssl

    def send_email(self, subject, sender, recipients, content):
        # TODO add crt arg

        msg = MIMEText(content)
        msg['Subject'] = subject
        msg['From'] = sender
        msg['To'] = ', '.join(recipients)

        # Send the message via the configured smtp server.
        # TODO keep same connection during some time ?
        s = None
        if self.ssl:
            s = smtplib.SMTP_SSL(self.host, self.port)
        else:
            s = smtplib.SMTP(self.host, self.port)

        if self.user and self.password:
            s.login(self.user, self.password)

        if self.start_tls:
            s.starttls()

        s.sendmail(sender, recipients, msg.as_string())

        s.quit()

    def process(self, msg):
        content = choose_first_not_none(self.content, msg.payload)
        subject = choose_first_not_none(self.subject, msg.meta.get('subject'), 'No subject')
        sender = choose_first_not_none(self.sender, msg.meta.get('sender'), 'pypeman@example.com')
        recipients = choose_first_not_none(self.recipients, msg.meta.get('recipients'), [])

        if isinstance(recipients, str):
            recipients = [recipients]

        self.send_email(subject, sender, recipients, content)

        return msg


class YielderNode(BaseNode):
    """
    Take an iterable msg.payload and returns a generator
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def process(self, msg):
        if not isinstance(msg.payload, collections.abc.Iterable):
            logger.error("Yielder node took a non iterable msg.payload: %r", msg.payload)
            raise Rejected()

        def generator(msg):
            payload = msg.payload
            ctx = msg.ctx
            meta = msg.meta
            for entry in payload:
                newmsg = Message(meta=meta, payload=entry)
                newmsg.ctx = ctx
                yield newmsg
        return generator(msg)


def reset_pypeman_nodes():
    """
    clears book keeping of all channels

    Can be useful for unit testing.
    """
    logger.info("clearing all_nodes and BaseNode._used-names.")
    all_nodes.clear()
    BaseNode._used_names.clear()

# Contrib nodes


# TODO: can we move this line to top of file?
from pypeman.helpers import lazyload  # noqa: E402

wrap = lazyload.Wrapper(__name__)

wrap.add_lazy('pypeman.contrib.ctx', "CombineCtx", [])
wrap.add_lazy('pypeman.contrib.xml', "XMLToPython", ["xmltodict"])
wrap.add_lazy('pypeman.contrib.xml', "PythonToXML", ["xmltodict"])
wrap.add_lazy('pypeman.contrib.hl7', "HL7ToPython", ["hl7"])
wrap.add_lazy('pypeman.contrib.hl7', "PythonToHL7", ["hl7"])
wrap.add_lazy('pypeman.contrib.http', "HttpRequest", ["aiohttp"])
wrap.add_lazy('pypeman.contrib.http', "RequestNode", ["aiohttp"])
wrap.add_lazy('pypeman.contrib.ftp', "FTPFileWriter", [])
wrap.add_lazy('pypeman.contrib.ftp', "FTPFileReader", [])
wrap.add_lazy('pypeman.contrib.ftp', "FTPFileDeleter", [])
wrap.add_lazy('pypeman.contrib.ftp', "FTPFileDeleter", [])
wrap.add_lazy('pypeman.contrib.csv', "CSV2Python", [])
wrap.add_lazy('pypeman.contrib.csv', "CSVstr2Python", [])
wrap.add_lazy('pypeman.contrib.csv', "Python2CSVstr", [])



import datetime

from pypeman import channels, message

from aiocron import crontab


class CronChannel(channels.BaseChannel):
    """
    This channel Launch processing at specified time interval.
    The first node of the channel receive a payload with the
    datetime of execution.

    :params cron: This set the interval. Accept any
        `aiocron <https://github.com/gawel/aiocron/>`_ compatible string.

    """

    def __init__(self, *args, cron='', **kwargs):
        super().__init__(*args, **kwargs)
        self.cron = cron

    async def start(self):
        await super().start()
        crontab(self.cron, func=self.tic, start=True, loop=self.loop)

    async def tic(self):
        msg = message.Message()
        msg.payload = datetime.datetime.now()
        await self.handle(msg)



import asyncio
import json
import os
import sys
import unittest

from unittest import mock
from pathlib import Path
from tempfile import TemporaryDirectory

import aiohttp
import pytest

from pypeman import nodes, message, conf, persistence

from pypeman.test import TearDownProjectTestCase as TestCase
from pypeman.tests.common import generate_msg
from pypeman.tests.common import LongNode


class FakeChannel():
    def __init__(self, loop):
        self.logger = mock.MagicMock()
        self.uuid = 'fakeChannel'
        self.name = 'fakeChannel'
        self.parent_uids = "parent_uid"
        self.parent_names = ["parent_names"]

        self.loop = loop


def tstfct(msg):
    return '/fctpath'


def tstfct2(msg):
    return 'fctname'


def get_mock_coro(return_value):

    async def mock_coro(*args, **kwargs):
        return return_value

    return mock.Mock(wraps=mock_coro)


class NodesTests(TestCase):
    def setUp(self):
        # Create class event loop used for tests to avoid failing
        # previous tests to impact next test ? (Not sure)
        self.loop = asyncio.new_event_loop()
        # Remove thread event loop to be sure we are not using
        # another event loop somewhere
        asyncio.set_event_loop(None)

    def test_base_node(self):
        """ if BaseNode() node functional """

        n = nodes.BaseNode()
        n.channel = FakeChannel(self.loop)

        m = generate_msg(message_content='test')

        ret = self.loop.run_until_complete(n.handle(m))

        # Check return
        self.assertTrue(isinstance(ret, message.Message))
        self.assertEqual(ret.payload, 'test', "Base node not working !")
        self.assertEqual(n.processed, 1, "Processed msg count broken")

    def test_unique_node_names(self):
        """ node names must be unique

            check that pypeman detects uniqeness violations
        """
        n = nodes.BaseNode(name="mynode")
        assert n is not None
        with pytest.raises(nodes.NodeException) as exc:
            m = nodes.BaseNode(name="mynode")
            assert m is not None
        assert "exists already" in str(exc._excinfo)

    def test_base_logging(self):
        """ whether BaseNode() node logging works"""

        n = nodes.BaseNode(log_output=True)
        n.channel = FakeChannel(self.loop)

        m = generate_msg(message_content='test')

        ret = self.loop.run_until_complete(n.handle(m))

        # Check return
        self.assertTrue(isinstance(ret, message.Message))
        self.assertEqual(ret.payload, 'test', "Base node not working !")
        self.assertEqual(n.processed, 1, "Processed msg count broken")

        n.channel.logger.log.assert_any_call(10, 'Payload: %r', 'test')
        n.channel.logger.log.assert_called_with(10, 'Meta: %r', {'question': 'unknown'})

    def test_node_persistence(self):
        """ Whether BaseNode() data memory persistence works"""
        persistence._backend = None # noqa

        # Mock settings
        with mock.patch('pypeman.persistence.settings',
                        conf.Settings(module_name='pypeman.tests.settings.test_settings_persistence')):
            result = []

            class TestPers(nodes.BaseNode):
                """ Test node for persistence """

                async def process(self, msg):
                    await self.save_data('test', 'value')
                    result.append(await self.restore_data('test'))
                    result.append(await self.restore_data('titi', default='Yo'))
                    try:
                        result.append(await self.restore_data('titi'))
                    except KeyError:
                        result.append('yay')

                    return msg

            n = TestPers()

            n.channel = FakeChannel(self.loop)

            m = generate_msg(message_content='test')

            self.loop.run_until_complete(n.handle(m))

            self.assertEqual(result[0], 'value', "Can't persist data for node")
            self.assertEqual(result[1], 'Yo', "Default value not working")
            self.assertEqual(result[2], 'yay', "Exception on missing key not working")

    def test_node_sqlite_persistence(self):
        """ Whether BaseNode() data sqlite persistence works"""
        persistence._backend = None # noqa

        db_path = '/tmp/to_be_removed_849827198746.sqlite'

        # Mock settings
        with mock.patch('pypeman.persistence.settings',
                        conf.Settings(module_name='pypeman.tests.settings.test_settings_sqlite_persist')):
            result = []

            class TestPers(nodes.BaseNode):
                """ Test node for persistence """

                async def process(self, msg):
                    await self.save_data('test', 'value')
                    result.append(await self.restore_data('test'))
                    result.append(await self.restore_data('titi', default='Yo'))
                    try:
                        result.append(await self.restore_data('titi'))
                    except KeyError:
                        result.append('yay')

                    return msg

            n = TestPers()

            n.channel = FakeChannel(self.loop)

            m = generate_msg(message_content='test')
            print('self', id(self.loop))

            self.loop.run_until_complete(n.handle(m))

            self.assertEqual(result[0], 'value', "Can't persist data for node")
            self.assertEqual(result[1], 'Yo', "Default value not working")
            self.assertEqual(result[2], 'yay', "Exception on missing key not working")

            os.remove(db_path)

    def test_log_node(self):
        """ whether Log() node functional """

        n = nodes.Log()
        n.channel = FakeChannel(self.loop)

        m = generate_msg()

        ret = self.loop.run_until_complete(n.handle(m))

        self.assertTrue(isinstance(ret, message.Message))

    def test_sleep_node(self):
        """ if Sleep() node functional """

        n = nodes.Sleep()
        n.channel = FakeChannel(self.loop)

        m = generate_msg(message_content='test')

        ret = self.loop.run_until_complete(n.handle(m))

        # Check return
        self.assertTrue(isinstance(ret, message.Message))
        self.assertEqual(ret.payload, 'test', "Sleep node not working !")

    def test_drop_node(self):
        """ Whether Drop() node is working """

        msg_to_show = "It's only dropped"
        n = nodes.Drop(message=msg_to_show)
        n.channel = FakeChannel(self.loop)

        m = generate_msg(message_content='test')

        with self.assertRaises(nodes.Dropped) as cm:
            self.loop.run_until_complete(n.handle(m))

        self.assertEqual(str(cm.exception), msg_to_show, "Drop node message not working !")

    def test_b64_nodes(self):
        """ if B64 nodes are functional """

        n1 = nodes.B64Encode()
        n2 = nodes.B64Decode()

        channel = FakeChannel(self.loop)

        n1.channel = channel
        n2.channel = channel

        m = generate_msg()

        m.payload = b'hello'

        base = bytes(m.payload)

        ret = self.loop.run_until_complete(n1.handle(m))
        ext_new = self.loop.run_until_complete(n2.handle(ret))

        # Check return
        self.assertTrue(isinstance(ret, message.Message))
        self.assertEqual(base, ext_new.payload, "B64 nodes not working !")

    def test_json_to_python_node(self):
        """ if JsonToPython() node functional """

        n = nodes.JsonToPython()
        n.channel = FakeChannel(self.loop)

        m = generate_msg(message_content='{"test":1}')

        ret = self.loop.run_until_complete(n.handle(m))
        # Check return
        self.assertTrue(isinstance(ret, message.Message))
        self.assertEqual(ret.payload, {"test": 1}, "JsonToPython node not working !")

    def test_thread_node(self):
        """ if Thread node functional """

        # TODO test if another task can be executed in //

        n = LongNode()
        n.channel = FakeChannel(self.loop)

        m = generate_msg()

        ret = self.loop.run_until_complete(n.handle(m))
        # Check return
        self.assertTrue(isinstance(ret, message.Message))

    @unittest.skipIf(not os.environ.get('PYPEMAN_TEST_SMTP_USER')
                     or not os.environ.get('PYPEMAN_TEST_SMTP_PASSWORD')
                     or not os.environ.get('PYPEMAN_TEST_RECIPIENT_EMAIL'),
                     "Email node test skipped. Set PYPEMAN_TEST_SMTP_USER, PYPEMAN_TEST_SMTP_PASSWORD and "
                     "PYPEMAN_TEST_RECIPIENT_EMAIL to enable it.")
    def test_email_node(self):
        """ if Email() node is functional """

        # Set this three vars to enable the skipped test.
        # Use mailtrap free account to test it.
        recipients = os.environ['PYPEMAN_TEST_RECIPIENT_EMAIL']
        smtp_user = os.environ['PYPEMAN_TEST_SMTP_USER']
        smtp_password = os.environ['PYPEMAN_TEST_SMTP_PASSWORD']

        n = nodes.Email(host="mailtrap.io", port=2525, start_tls=False,
                        ssl=False, user=smtp_user, password=smtp_password,
                        subject="Sent from email node of Pypeman",
                        sender=recipients, recipients=recipients)
        n.channel = FakeChannel(self.loop)

        m = generate_msg()
        m.payload = "Message content is full of silence !"

        ret = self.loop.run_until_complete(n.handle(m))

        self.assertTrue(isinstance(ret, message.Message))

    def test_save_node(self):
        """ if Save() node functional """

        with mock.patch("builtins.open", mock.mock_open(read_data="data")) as mock_file, \
                mock.patch('pypeman.nodes.os.makedirs') as mock_makedirs:

            mock_makedirs.return_value = None

            n = nodes.Save(uri='file:///tmp/test/?filename=%(msg_year)s/'
                           '%(msg_month)s/message%(msg_day)s-%(counter)s.txt')
            n.channel = FakeChannel(self.loop)

            m = generate_msg(timestamp=(1981, 12, 28, 13, 37))
            m.payload = "content"

            ret = self.loop.run_until_complete(n.handle(m))

            self.assertTrue(isinstance(ret, message.Message))

            # Asserts
            mock_makedirs.assert_called_once_with('/tmp/test/1981/12')
            mock_file.assert_called_once_with('/tmp/test/1981/12/message28-0.txt', 'w')
            handle = mock_file()
            handle.write.assert_called_once_with('content')

    def test_yielder_node(self):
        """ Whether YielderNode is functional """
        yieldernode = nodes.YielderNode()
        data = [1, 2, 3]
        msg = message.Message(payload=data)
        out_gen = yieldernode.process(msg)
        idx = 0
        for entry in out_gen:
            self.assertEqual(data[idx], entry.payload)
            idx += 1

    def test_xml_nodes(self):
        """ if XML nodes are functional """
        try:
            import xmltodict  # noqa F401
        except ImportError:
            raise unittest.SkipTest("Missing dependency xmltodict.")

        n1 = nodes.XMLToPython()
        n2 = nodes.PythonToXML()

        channel = FakeChannel(self.loop)

        n1.channel = channel
        n2.channel = channel

        m = generate_msg()

        m.payload = '<?xml version="1.0" encoding="utf-8"?>\n<test>hello</test>'

        base = str(m.payload)

        ret = self.loop.run_until_complete(n1.handle(m))
        ext_new = self.loop.run_until_complete(n2.handle(ret))
        # Check return
        self.assertTrue(isinstance(ret, message.Message))
        self.assertEqual(base, ext_new.payload, "XML nodes not working !")

    def test_ftp_nodes(self):
        """ Whether FTP nodes are functional """

        channel = FakeChannel(self.loop)

        ftp_config = dict(host="fake", credentials=("fake", "fake"))

        fake_ftp = mock.MagicMock()
        fake_ftp.download_file = mock.Mock(return_value=b"new_content")

        fake_ftp_helper = mock.Mock(return_value=fake_ftp)

        with mock.patch('pypeman.contrib.ftp.FTPHelper', new=fake_ftp_helper):

            reader = nodes.FTPFileReader(filepath="test_read", **ftp_config)
            delete = nodes.FTPFileDeleter(filepath="test_delete", **ftp_config)

            writer = nodes.FTPFileWriter(filepath="test_write", **ftp_config)

            reader.channel = channel
            delete.channel = channel
            writer.channel = channel

            m1 = generate_msg(message_content="to_be_replaced")
            m1_delete = generate_msg(message_content="to_be_replaced")
            m2 = generate_msg(message_content="message_content")

            # Test reader
            result = self.loop.run_until_complete(reader.handle(m1))

            fake_ftp.download_file.assert_called_once_with('test_read')
            self.assertEqual(result.payload, b"new_content", "FTP reader not working")

            # Test reader with delete after
            result = self.loop.run_until_complete(delete.handle(m1_delete))

            fake_ftp.delete.assert_called_once_with('test_delete')

            # test writer
            result = self.loop.run_until_complete(writer.handle(m2))
            fake_ftp.upload_file.assert_called_once_with('test_write.part', 'message_content')
            fake_ftp.rename.assert_called_once_with('test_write.part', 'test_write')

    @unittest.skipIf((sys.version_info[:2] == (3, 7)),
                     "difficulty to mock async with statement in py3.7")  # TODO: rm in py3.8+
    def test_httprequest_node(self):
        """ Whether HttpRequest node is functional """

        channel = FakeChannel(self.loop)

        auth = ("login", "mdp")
        url = 'http://url/%(meta.beta)s/%(payload.alpha)s'
        b_auth = aiohttp.BasicAuth(*auth)
        client_cert = ('/cert.key', '/cert.crt')
        http_node1 = nodes.HttpRequest(url=url, verify=False, auth=auth)
        http_node1.channel = channel

        content1 = {"alpha": "payload_url"}
        msg1 = generate_msg(message_content=content1)
        meta_params = {'omega': 'meta_params'}
        headers1 = {'test': 'test'}
        msg1.meta = {
            "beta": "meta_url", 'params': meta_params, 'headers': headers1}
        req_url1 = 'http://url/meta_url/payload_url'
        req_kwargs1 = {
            'data': None,
            'params': [('omega', 'meta_params')],
            'url': req_url1,
            'headers': headers1,
            'method': 'get',
            'auth': b_auth
            }

        msg2 = generate_msg(message_content=content1)
        msg2.meta = dict(msg1.meta)
        msg2.meta['method'] = 'post'
        msg2.meta['params'] = {'zeta': ['un', 'deux', 'trois']}
        req_kwargs2 = dict(req_kwargs1)
        req_kwargs2['method'] = 'post'
        req_kwargs2['params'] = [
                ('zeta', 'un'),
                ('zeta', 'deux'),
                ('zeta', 'trois'),
                ]
        req_kwargs2['data'] = content1

        args_headers = {'args_headers': 'args_headers'}
        args_params = {'theta': ['uno', 'dos'], 'omega': tstfct2}
        http_node2 = nodes.HttpRequest(
            url=url,
            method='post',
            client_cert=client_cert,
            auth=b_auth,
            headers=args_headers,
            params=args_params
        )
        http_node2.channel = channel
        http_node3 = nodes.HttpRequest(
            url=url,
            verify=False,
            method='get',
            binary=True,
        )
        http_node3.channel = channel
        http_node4 = nodes.HttpRequest(
            url=url,
            method='post',
            send_as_json=True,
            headers=args_headers,
            params=args_params
        )
        http_node4.channel = channel
        msg3 = msg1.copy()
        req_kwargs3 = dict(req_kwargs1)
        req_kwargs3['method'] = 'post'
        req_kwargs3['params'] = [
                ('theta', 'uno'),
                ('theta', 'dos'),
                ('omega', 'fctname'),
                ]
        req_kwargs3['headers'] = args_headers
        req_kwargs3['data'] = content1
        msg4 = msg1.copy()
        req_kwargs4 = dict(req_kwargs3)
        req_kwargs4['json'] = req_kwargs4.pop("data")
        req_kwargs4['auth'] = None
        msg5 = msg1.copy()

        with mock.patch(
            'pypeman.contrib.http.aiohttp.ClientSession',
            autospec=True) as mock_client_session, mock.patch(
                'ssl.SSLContext',
                autospec=True) as mock_ssl_context:
            mock_ctx_mgr = mock_client_session.return_value
            mock_session = mock_ctx_mgr.__aenter__.return_value
            mg = mock.MagicMock()
            mg.text = get_mock_coro(mock.MagicMock())
            mock_session.request = get_mock_coro(mg)
            mock_load_cert_chain = mock_ssl_context.return_value.load_cert_chain

            """
                Test 1:
                - default get,
                - auth tuple in object BasicAuth,
                - simple dict params from meta,
                - headers from meta
                - url construction
            """
            self.loop.run_until_complete(http_node1.handle(msg1))
            mock_session.request.assert_called_once_with(**req_kwargs1)
            mock_load_cert_chain.assert_not_called()

            mock_session.reset_mock()

            """
                Test 2:
                - post in meta with data from content,
                - list in dict params from meta,
            """
            self.loop.run_until_complete(http_node1.handle(msg2))
            mock_session.request.assert_called_once_with(**req_kwargs2)
            mock_load_cert_chain.assert_not_called()

            mock_session.reset_mock()

            """
                Test 3:
                - post in node args,
                - object BasicAuth for auth,
                - list in dict params from args + callable str param,
                - headers from args
                - client_cert
            """
            self.loop.run_until_complete(http_node2.handle(msg3))
            mock_session.request.assert_called_once_with(**req_kwargs3)
            mock_load_cert_chain.assert_called_once_with(client_cert[0], client_cert[1])
            mock_session.reset_mock()

            """
                Test 4:
                - get bytes content
            """
            byte_msg = bytes("coucou", "utf-8")
            mg = mock.MagicMock()
            mg.read = get_mock_coro(byte_msg)
            mock_session.request = get_mock_coro(mg)
            outmsg = self.loop.run_until_complete(http_node3.handle(msg4))
            mock_session.request.assert_called_once()
            mg.read.assert_called_once()
            self.assertEqual(outmsg.payload, byte_msg)
            mock_session.reset_mock()

            """
                    Test 5:
                    - get json content
            """
            # channel = FakeChannel(self.loop)
            url = 'http://url/titi/tata'
            http_jsonnode = nodes.HttpRequest(url=url, verify=False, json=True)
            http_jsonnode.channel = channel
            data = {"titi": "tata"}
            jsondata = json.dumps(data)
            emptymsg = message.Message()
            mg = mock.MagicMock()
            mg.text = get_mock_coro(jsondata)
            mock_session.request = get_mock_coro(mg)
            outdata = self.loop.run_until_complete(http_jsonnode.process(emptymsg))
            self.assertEqual(data, outdata.payload)

            """
                Test 6:
                - post json with data from content
            """
            mock_session.reset_mock()
            mock_load_cert_chain.reset_mock()
            self.loop.run_until_complete(http_node4.handle(msg5))
            mock_session.request.assert_called_once_with(**req_kwargs4)
            mock_load_cert_chain.assert_not_called()

            mock_session.reset_mock()

    @unittest.skipIf((sys.version_info[:2] == (3, 7)),
                     "difficulty to mock async with statement in py3.7")  # TODO: rm in py3.8+
    def test_httprequest_node2_new_parsing(self):
        """ Whether HttpRequest node recursive url parser is functional """

        channel = FakeChannel(self.loop)

        auth = ("login", "mdp")
        url = 'http://url/%(meta.beta.beta2)s/%(payload.alpha.toto)s'
        b_auth = aiohttp.BasicAuth(*auth)
        client_cert = ('/cert.key', '/cert.crt')
        http_node1 = nodes.HttpRequest(
            url=url, verify=False, auth=auth,
            old_url_parsing=False,)
        http_node1.channel = channel

        content1 = {"alpha": {"toto": "payload_url"}}
        msg1 = generate_msg(message_content=content1)
        meta_params = {'omega': 'meta_params'}
        headers1 = {'test': 'test'}
        msg1.meta = {
            "beta": {"beta2": "meta_url"}, 'params': meta_params, 'headers': headers1}
        req_url1 = 'http://url/meta_url/payload_url'
        req_kwargs1 = {
            'data': None,
            'params': [('omega', 'meta_params')],
            'url': req_url1,
            'headers': headers1,
            'method': 'get',
            'auth': b_auth
            }

        msg2 = generate_msg(message_content=content1)
        msg2.meta = dict(msg1.meta)
        msg2.meta['method'] = 'post'
        msg2.meta['params'] = {'zeta': ['un', 'deux', 'trois']}
        req_kwargs2 = dict(req_kwargs1)
        req_kwargs2['method'] = 'post'
        req_kwargs2['params'] = [
                ('zeta', 'un'),
                ('zeta', 'deux'),
                ('zeta', 'trois'),
                ]
        req_kwargs2['data'] = content1

        args_headers = {'args_headers': 'args_headers'}
        args_params = {'theta': ['uno', 'dos'], 'omega': tstfct2}
        http_node2 = nodes.HttpRequest(
            url=url,
            method='post',
            client_cert=client_cert,
            auth=b_auth,
            headers=args_headers,
            params=args_params,
            old_url_parsing=False,
        )
        http_node2.channel = channel

        with mock.patch(
            'pypeman.contrib.http.aiohttp.ClientSession',
            autospec=True) as mock_client_session, mock.patch(
                'ssl.SSLContext',
                autospec=True) as mock_ssl_context:
            mock_ctx_mgr = mock_client_session.return_value
            mock_session = mock_ctx_mgr.__aenter__.return_value
            mg = mock.MagicMock()
            mg.text = get_mock_coro(mock.MagicMock())
            mock_session.request = get_mock_coro(mg)
            mock_load_cert_chain = mock_ssl_context.return_value.load_cert_chain

            """
                Test 1:
                - default get,
                - auth tuple in object BasicAuth,
                - imbricated dict params from meta,
                - headers from meta
                - url construction
            """
            self.loop.run_until_complete(http_node1.handle(msg1))
            mock_session.request.assert_called_once_with(**req_kwargs1)
            mock_load_cert_chain.assert_not_called()

            mock_session.reset_mock()

            """
                Test 2:
                - post in meta with data from content,
                - imbricated dict params from meta,
            """
            self.loop.run_until_complete(http_node1.handle(msg2))
            mock_session.request.assert_called_once_with(**req_kwargs2)
            mock_load_cert_chain.assert_not_called()

            mock_session.reset_mock()

    def test_file_reader_node(self):
        """if FileReader are functionnal"""

        reader = nodes.FileReader(filepath='/filepath', filename='badname')
        channel = FakeChannel(self.loop)

        reader.channel = channel
        msg1 = generate_msg()

        with mock.patch("builtins.open", mock.mock_open(read_data="data")) as mock_file:
            result = self.loop.run_until_complete(reader.handle(msg1))

        mock_file.assert_called_once_with('/filepath', 'r')
        self.assertEqual(result.payload, "data", "FileReader not working")

        reader2 = nodes.FileReader()
        reader2.channel = channel
        msg2 = generate_msg()
        msg2.meta['filepath'] = '/filepath2'
        msg2.meta['filename'] = '/badpath'

        with mock.patch("builtins.open", mock.mock_open(read_data="data2")) as mock_file:
            result = self.loop.run_until_complete(reader2.handle(msg2))

        mock_file.assert_called_once_with('/filepath2', 'r')
        self.assertEqual(result.payload, "data2", "FileReader not working with meta")

        reader3 = nodes.FileReader(filepath=tstfct, filename='badname')
        reader3.channel = channel

        msg3 = generate_msg()
        msg3.meta['filepath'] = '/badpath'
        msg3.meta['filename'] = 'badname2'

        with mock.patch("builtins.open", mock.mock_open(read_data="data")) as mock_file:
            result = self.loop.run_until_complete(reader3.handle(msg3))

        mock_file.assert_called_once_with('/fctpath', 'r')

        reader4 = nodes.FileReader(filename=tstfct2)
        reader4.channel = channel
        msg4 = generate_msg()
        msg4.meta['filepath'] = '/filepath3/badname'
        msg4.meta['filename'] = 'badname'

        with mock.patch("builtins.open", mock.mock_open(read_data="data")) as mock_file:
            result = self.loop.run_until_complete(reader4.handle(msg4))

        mock_file.assert_called_once_with('/filepath3/fctname', 'r')

    def test_file_writer_node(self):
        """Whether FileWriter is functionnal"""
        writer = nodes.FileWriter(filepath='/filepath', safe_file=False)
        channel = FakeChannel(self.loop)
        writer.channel = channel
        msg1 = generate_msg(message_content="message_content")
        with mock.patch("builtins.open", mock.mock_open()) as mock_file:
            self.loop.run_until_complete(writer.handle(msg1))

        mock_file.assert_called_once_with('/filepath', 'w')
        handle = mock_file()
        handle.write.assert_called_once_with('message_content')

        writer2 = nodes.FileWriter(safe_file=False)
        writer2.channel = channel
        msg2 = generate_msg(message_content="message_content2")
        msg2.meta['filepath'] = '/filepath2'
        with mock.patch("builtins.open", mock.mock_open()) as mock_file:
            self.loop.run_until_complete(writer2.handle(msg2))

        mock_file.assert_called_once_with('/filepath2', 'w')
        handle = mock_file()
        handle.write.assert_called_once_with('message_content2')


class TestsFileNodes(TestCase):
    """ Tests common.com_nodes File Nodes
    """
    tmpdir_obj = None

    @classmethod
    def setUpClass(self):
        self.tmpdir_obj = TemporaryDirectory()

    @classmethod
    def tearDownClass(self):
        self.tmpdir_obj.cleanup()
        self.tmpdir_obj = None

    def test_filemover(self):
        indir_pth = Path(self.tmpdir_obj.name)
        outdir_pth = indir_pth / "out"
        fname = "test_fmover.txt"
        src_fpath_pth = indir_pth / fname
        dst_fpath_pth = outdir_pth / fname
        src_fpath_pth.touch()
        fmover_node = nodes.FileMover(dest_path=outdir_pth)
        msg = message.Message(meta={"filename": fname, "filepath": str(src_fpath_pth)})
        fmover_node.process(msg)
        self.assertFalse(src_fpath_pth.exists(), msg="file not moved (already in indir)")
        self.assertTrue(dst_fpath_pth.exists(), msg="file not moved (not in outdir)")
        return msg

    def test_filecleaner(self):
        indir_pth = Path(self.tmpdir_obj.name)
        fname = "test_fcleaner.txt"
        fname2 = "test_fcleaner.rm"
        src_fpath_pth = indir_pth / fname
        src2_fpath_pth = indir_pth / fname2
        src_fpath_pth.touch()
        src2_fpath_pth.touch()
        fcleaner_node = nodes.FileCleaner(extensions_to_rm=[".rm"])
        msg = message.Message(meta={"filename": fname, "filepath": str(src_fpath_pth)})
        fcleaner_node.process(msg)
        self.assertFalse(src_fpath_pth.exists(), msg="file 1 not rm")
        self.assertFalse(src2_fpath_pth.exists(), msg="file 2 not rm")
        return msg


class TestsCsvContrib(TestCase):
    """ Tests common.com_nodes File Nodes
    """
    tmpdir_obj = None
    py_data_dict = [
        {"id": "1", "msg": "msg1", "ty": "ty1"},
        {"id": "2", "msg": "msg2", "ty": "ty2"},
        {"id": "3", "msg": "msg3", "ty": "ty3"},
    ]
    py_data_nodict = [
        ["id", "msg", "ty"],
        ["1", "msg1", "ty1"],
        ["2", "msg2", "ty2"],
        ["3", "msg3", "ty3"],
    ]
    csv_data_fpath = Path(__file__).parent / "data" / "csv_test_data.csv"
    with open(csv_data_fpath, "r", newline="") as fin:
        csv_str_data = fin.read()

    def test_csv2python(self):
        csv2py_node = nodes.CSV2Python(to_dict=True, headers=True)
        csv2py_node2 = nodes.CSV2Python(to_dict=False, headers=False)
        msg = message.Message(meta={"filepath": self.csv_data_fpath})
        msg2 = msg.copy()
        # Test with to_dict and headers params activated
        processed_msg = csv2py_node.process(msg)
        self.assertListEqual(self.py_data_dict, processed_msg.payload)
        # Test without to_dict and headers params
        processed_msg2 = csv2py_node2.process(msg2)
        self.assertListEqual(self.py_data_nodict, processed_msg2.payload)
        return msg

    def test_csvstr2python(self):
        csv2py_node = nodes.CSVstr2Python(to_dict=True, headers=True)
        csv2py_node2 = nodes.CSVstr2Python(to_dict=False, headers=False)
        msg = message.Message(payload=self.csv_str_data)
        msg2 = msg.copy()
        # Test with to_dict and headers params activated
        processed_msg = csv2py_node.process(msg)
        self.assertListEqual(self.py_data_dict, processed_msg.payload)
        # Test without to_dict and headers params
        processed_msg2 = csv2py_node2.process(msg2)
        self.assertListEqual(self.py_data_nodict, processed_msg2.payload)
        return msg

    def test_python2csvstr(self):
        py2csv_node = nodes.Python2CSVstr(header=True)
        msg = message.Message(payload=self.py_data_dict)
        processed_msg = py2csv_node.process(msg)
        self.assertEqual(self.csv_str_data, processed_msg.payload)
        return msg



#!/usr/bin/env python
import datetime
import uuid
import copy
from uuid import UUID
import pickle
import json
import base64
import logging

default_logger = logging.getLogger(__name__)

DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


class Message():
    """
        A message is the unity of informations exchanged between nodes of a
        channel.

        A message have following properties:

            :attribute payload: The message content.
            :attribute meta: The message metadata.
            :attribute timestamp: The creation date of message
            :attribute uuid: uuid to identify message
            :attribute content_type: Used ?
            :attribute ctx: Current context when you want to save a message for later use.

        :param payload: You can initialise the payload by setting this param.
        :param meta: Same as payload, you can initialise the meta by setting this param.

    """
    # TODO : add_ctx and delete_ctx

    PENDING = "pending"
    PROCESSING = "processing"
    ERROR = "error"
    REJECTED = "rejected"
    PROCESSED = "processed"

    def __init__(self, content_type='application/text', payload=None, meta=None):
        self.content_type = content_type
        self.timestamp = datetime.datetime.now()
        self.uuid = uuid.uuid4().hex

        self.payload = payload

        if meta is None:
            meta = {}
        self.meta = meta

        self.ctx = {}

    def timestamp_str(self):
        """ Return timestamp formated string """
        return self.timestamp.strftime(DATE_FORMAT)

    def copy(self):
        """
        Copy the message. Useful for channel forking purpose.

        :return: A copy of current message.
        """
        return copy.deepcopy(self)

    def renew(self):
        """
        Copy the message but update the `timestamp` and `uuid`.

        :return: A copy of current message with new `uuid` and `Timestamp`.
        """
        msg = self.copy()

        msg.uuid = uuid.uuid4().hex
        msg.timestamp = datetime.datetime.now()
        return msg

    def add_context(self, key, msg):
        """
        Add a msg to the `.ctx` property with specified key.

        :param key: Key to store message.
        :param msg: Message to store.
        """
        self.ctx[key] = dict(
            meta=dict(msg.meta),
            payload=copy.deepcopy(msg.payload),
        )

    def to_dict(self, encode_payload=True):
        """
        Convert the current message object to a dict.
        Payload is pickled and b64 encoded if encode_payload not set to False

        :return: A dict with an equivalent of message
        """
        result = {}
        result['timestamp'] = self.timestamp.strftime(DATE_FORMAT)
        result['uuid'] = self.uuid
        if encode_payload:
            result['payload'] = base64.b64encode(pickle.dumps(self.payload)).decode('ascii')
        else:
            try:
                result['payload'] = str(self.payload)
            except Exception:
                default_logger.warning("Cannot convert to string payload %r, pickling it")
                result['payload'] = base64.b64encode(pickle.dumps(self.payload)).decode('ascii')
        result['meta'] = self.meta
        result['ctx'] = {}

        for k, ctx_msg in self.ctx.items():
            result['ctx'][k] = {}
            result['ctx'][k]['payload'] = base64.b64encode(pickle.dumps(ctx_msg['payload'])).decode('ascii')
            result['ctx'][k]['meta'] = dict(ctx_msg['meta'])

        return result

    def to_json(self):
        """
        Create json string for current message.

        :return: a json string equivalent for message.
        """
        return json.dumps(self.to_dict())

    @staticmethod
    def from_dict(data):
        """
        Convert the input dict previously converted with `.as_dict()` method in Message object.

        :param data: The input dict.
        :return: The message message object correponding to given data.
        """
        result = Message()
        result.timestamp = datetime.datetime.strptime(data['timestamp'], DATE_FORMAT)
        result.uuid = UUID(data['uuid']).hex
        result.payload = pickle.loads(base64.b64decode(data['payload'].encode('ascii')))
        result.meta = data['meta']

        for k, ctx_msg in data['ctx'].items():
            result.ctx[k] = {}
            result.ctx[k]['payload'] = pickle.loads(base64.b64decode(ctx_msg['payload'].encode('ascii')))
            result.ctx[k]['meta'] = dict(ctx_msg['meta'])

        return result

    @staticmethod
    def from_json(data):
        """
        Create a message from previously saved json string.

        :param data: Data to read message from.
        :return: A new message instance created from json data.
        """
        msg = Message.from_dict(json.loads(data))
        return msg

    def log(self, logger=default_logger, log_level=logging.DEBUG, payload=True, meta=True, context=False):
        """
        Log a message.

        :param logger: Logger
        :param log_level: log level for all log.
        :param payload: Whether log payload.
        :param meta: Whether log meta.
        :param context: Whether log context.
        """

        if payload:
            logger.log(log_level, 'Payload: %r', self.payload)

        if meta:
            logger.log(log_level, 'Meta: %r', self.meta)

        if context and self.ctx:
            logger.log(log_level, 'Context for message ->')
            for key, msg in self.ctx.items():
                logger.log(log_level, '-- Key "%s" --', key)
                if payload:
                    logger.log(log_level, 'Payload: %r', msg['payload'])

                if meta:
                    logger.log(log_level, 'Meta: %r', msg['meta'])

    def to_print(self, payload=True, meta=True, context=False):
        """
        Return a printable version of message.

        :param payload: Whether print payload.
        :param meta: Whether print meta.
        :param context: Whether print context.
        """

        result = "Message {msg.uuid}\nDate: {msg.timestamp}\n".format(msg=self)

        if payload:
            result += 'Payload: %r\n' % self.payload

        if meta:
            result += 'Meta: %r\n' % self.meta

        if context and self.ctx:
            result += 'Context for message ->\n'
            for key, msg in self.ctx.items():
                result += '-- Key "%s" --\n' % key
                if payload:
                    result += 'Payload: %r\n' % msg['payload']

                if meta:
                    result += 'Meta: %r\n' % msg['meta']

        return result

    def __str__(self):
        return "<msg: %s>" % self.uuid

    def __repr__(self):
        return "<msg: %s- id(%s)>" % (self.uuid, id(self))



from pypeman.default_settings import *  # noqa F403


handlers = LOGGING['handlers']  # noqa F405
loggers = LOGGING['loggers']  # noqa F405

handlers['debug'] = {
    'class': 'pypeman.helpers.logging.DebugLogHandler',
    'level': 'DEBUG',
    }

loggers['']['level'] = 'DEBUG'



""" helpers for logging
"""

import logging
log_trace = []


class DebugLogHandler(logging.Handler):
    """ a small debug logger just storing all records in a list """
    def __init__(self):
        super().__init__()
        self.log_trace = []

    def emit(self, record):
        self.log_trace.append(record)

    def num_entries(self):
        return len(self.log_trace)

    def show_entries(self):
        print("Got %d entries" % len(self.log_trace))
        for idx, rec in enumerate(self.log_trace):
            print("%2d %8.3f %s:%d %r" % (idx, rec.relativeCreated/1000.,
                  rec.pathname,  rec.lineno,
                  rec.message))



import asyncio
import logging
import re

from ftplib import FTP
from io import BytesIO
from pathlib import Path

from concurrent.futures import ThreadPoolExecutor

from pypeman import channels, nodes, message

# Can be redefined
default_thread_pool = ThreadPoolExecutor(max_workers=3)

logger = logging.getLogger(__name__)


class FTPConnection():
    """
    FTP connection manager.
    """
    def __init__(self, host, port, credentials):
        """
        :param host: FTP host.
        :param port: FTP port
        :param credentials: A tuple with (login, password,)
        :return:
        """
        self.host = host
        self.port = port
        self.credentials = credentials

    def __enter__(self):
        self.ftp = FTP(self.host)
        self.ftp.login(*self.credentials)
        return self.ftp

    def __exit__(self, type, value, tb):
        try:
            self.ftp.quit()
        except Exception:
            self.ftp.close()


class FTPHelper():
    """
    FTP helper to abstract ftp access.
    """
    # TODO reuse ftp connection if possible

    def __init__(self, host, port, credentials):
        self.host = host
        self.port = port
        self.credentials = credentials

    def list_dir(self, path):
        with FTPConnection(self.host, self.port, self.credentials) as ftp_conn:
            ftp_conn.cwd(path)
            return set(ftp_conn.nlst())

    def download_file(self, filepath):
        """
        Download a file from ftp asynchronously.
        :param filepath: file path to download.
        :return: content of the file.
        """
        output = BytesIO()

        # Get file from ftp
        with FTPConnection(self.host, self.port, self.credentials) as ftp_conn:
            ftp_conn.retrbinary("RETR %s" % filepath, output.write)

        content = output.getvalue()
        output.close()

        return content

    def upload_file(self, filepath, content):
        """
        Upload an file to ftp.
        :param filepath: Path of file to create.
        :param content: Content to upload.
        """
        input = BytesIO(content)

        # Get file from ftp
        with FTPConnection(self.host, self.port, self.credentials) as ftp_conn:
            ftp_conn.storbinary('STOR %s' % filepath, input)

        input.close()

        return content

    def rename(self, fromfilepath, tofilepath):
        """
        Rename a file from path to another path in ftp.
        :param fromfilepath: original file to rename.
        :param tofilepath: destination file.
        """
        with FTPConnection(self.host, self.port, self.credentials) as ftp_conn:
            ftp_conn.rename(fromfilepath, tofilepath)

    def delete(self, filepath):
        """
        Delete an FTP file.
        :param filepath: File to delete.
        """
        with FTPConnection(self.host, self.port, self.credentials) as ftp_conn:
            ftp_conn.delete(filepath)


class FTPWatcherChannel(channels.BaseChannel):
    """
    Channel that watch ftp for file creation.
    """
    def __init__(self, *args, host="", port=21, credentials="", basedir="", regex='.*',
                 interval=60, delete_after=False, encoding="utf-8",
                 thread_pool=None, sort_function=sorted, real_extensions=None,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self.basedir = basedir
        self.interval = interval
        self.delete_after = delete_after
        self.re = re.compile(regex)
        self.encoding = encoding
        self.sort_function = sort_function
        self.real_extensions = real_extensions  # list of extensions for exemple: [".csv", ".CSV"]
        self.ls_prev = set()

        # pool used to make ftp connection out of thread
        self.executor = thread_pool or default_thread_pool

        self.ftphelper = FTPHelper(host, port, credentials)

    async def start(self):
        await super().start()
        asyncio.create_task(self.watch_for_file())

    def download_file(self, filename):
        """
        Download a file from ftp asynchronously.

        :param filepath: file path to download.

        :return: Content of the downloaded file.
        """
        if not self.is_stopped():
            return self.ftphelper.download_file(self.basedir + '/' + filename)

    async def get_file_and_process(self, filename):
        """
        Download a file from ftp and launch channel processing on msg with result as payload.
        Also add a `filepath` header with ftp relative path of downloaded file.

        :param filename: file to download relative to `basedir`.

        :return: processed result
        """

        payload = await self.loop.run_in_executor(self.executor, self.download_file, filename)

        msg = message.Message()
        msg.payload = payload
        msg.meta['filepath'] = self.basedir + '/' + filename

        if not self.is_stopped():
            result = await self.handle(msg)

            if self.delete_after:
                await self.loop.run_in_executor(
                                self.executor,
                                self.ftphelper.delete,
                                self.basedir + '/' + filename)

            return result

    async def tick(self):
        """
        One iteration of watching.
        """

        ls = await self.loop.run_in_executor(self.executor, self.ftphelper.list_dir, self.basedir)

        # Make diff from previous one.
        added = self.sort_function(ls-self.ls_prev)
        self.ls_prev = ls

        for filename in added:
            if self.re.match(filename) and not self.is_stopped():
                if self.real_extensions:
                    for extension in self.real_extensions:
                        real_fname = str(Path(filename).with_suffix(extension))
                        if real_fname in ls:
                            filename = real_fname
                            break
                    else:
                        # If no related files
                        # TODO : ask if raise exc or not
                        logger.error(
                            "No %r related file to %s",
                            self.real_extensions, filename)
                        continue
                asyncio.create_task(self.get_file_and_process(filename))

    async def watch_for_file(self):
        """
        Watch recursively for ftp new files.
        If file match regex, it is downloaded then processed in a message.
        """
        await asyncio.sleep(self.interval)
        try:
            await self.tick()
        finally:
            if not self.is_stopped():
                asyncio.create_task(self.watch_for_file())


class FTPFileReader(nodes.ThreadNode):
    """
    Node to read a file from FTP.
    """
    def __init__(self, host="", port=21, credentials=None, filepath=None, **kwargs):

        super().__init__(**kwargs)

        self.filepath = filepath

        self.ftphelper = FTPHelper(host, port, credentials)

    def process(self, msg):

        filepath = nodes.choose_first_not_none(
            nodes.callable_or_value(self.filepath, msg),
            msg.meta.get('filepath'))

        content = self.ftphelper.download_file(filepath)

        msg.payload = content
        msg.meta['filepath'] = filepath

        return msg


class FTPFileDeleter(nodes.ThreadNode):
    """
    Node to delete a file from FTP.
    """
    def __init__(self, host="", port=21, credentials=None, filepath=None, **kwargs):

        super().__init__(**kwargs)

        self.filepath = filepath

        self.ftphelper = FTPHelper(host, port, credentials)

    def process(self, msg):

        filepath = nodes.choose_first_not_none(
            nodes.callable_or_value(self.filepath, msg),
            msg.meta.get('filepath'))

        self.ftphelper.delete(filepath)

        return msg


class FTPFileWriter(nodes.ThreadNode):
    """
    Node to write content to FTP. File is first written with `.part` concatenated
    to its name then renamed to avoid partial upload.
    """
    def __init__(self, host="", port=21, credentials=None, filepath=None, **kwargs):

        super().__init__(**kwargs)

        self.filepath = filepath

        self.ftphelper = FTPHelper(host, port, credentials)

    def process(self, msg):

        filepath = nodes.choose_first_not_none(
            nodes.callable_or_value(self.filepath, msg),
            msg.meta.get('filepath'))

        self.ftphelper.upload_file(filepath + '.part', msg.payload)
        self.ftphelper.rename(filepath + '.part', filepath)

        return msg



import argparse
import ast
import asyncio
import json

from pathlib import Path

from pypeman.msgstore import FileMessageStore


class Filter:
    """
    a simple filter to filter messages fetched from a storage
    """
    def __init__(self, filter_str):
        self.filter_str = filter_str
        name, val = filter_str.split("=", 1)
        self.name = name
        self.val = ast.literal_eval(val)

    def __str__(self):
        return self.filter_str

    def match(self, msg):
        return msg.payload.get(self.name) == self.val

    def __repr__(self):
        return str(self)


def split_path_to_store_n_id(path):
    """
    helper to split an absolute msg_store path to a store and its id-path
    """
    path_wo_date = path.parents[3]
    store_name = path_wo_date.name
    store_path = path_wo_date.parent
    # TODO: most refactor the stores to simplify access by uid or file name or bu
    # TODO: absolute file name
    store = FileMessageStore(str(store_path), store_name)
    rel_path_w_date = path.relative_to(path_wo_date)
    return store, rel_path_w_date


def print_msg(store_id, msg):
    """
    just print store id and msg payload
    """
    print(store_id, msg.payload)


def handle_store_entry(path, pl_type=None, filters=None, action=print):
    """
    handles a store entry and print if not rejected by filters
    """
    store, rel_path_w_date = split_path_to_store_n_id(path)
    entry = asyncio.run(store.get(str(rel_path_w_date)))
    msg = entry["message"]
    msg_id = entry["id"]
    payload = msg.payload
    if pl_type == "json":
        msg.payload = payload = json.loads(payload)

    match = True
    for filter in filters:
        if not filter.match(msg):
            match = False
            break
    if not match:
        return
    action(msg_id, msg)


def process_file_store(
        path, pl_type=None, filters=None, action=print_msg):
    """
    handles one or multiple file store messages
    must later see how to make code handle also other stores

    param: path: the store 'id' / path of the store entry
    param: filters: list of filters to filter message objects
    param: action: action to perform with entry
    """

    if path.is_file():
        handle_store_entry(
            path, pl_type=pl_type, filters=filters, action=action)

    all_entries = path.glob("**/*")
    f_paths = (
        path for path in all_entries
        if path.is_file() and not path.suffix
    )
    for entry in f_paths:
        handle_store_entry(
            entry, pl_type=pl_type, filters=filters, action=action)


def mk_parser():
    """
    argument parser for view_store command
    """
    descr = "view store contents or search"
    parser = argparse.ArgumentParser(descr)
    parser.add_argument("path", help="path to search messages")
    parser.add_argument(
        "--type", "-t",
        default="json",
        help="payload type (dflt = %(default)s)",
    )
    parser.add_argument(
        "--filter", "-f",
        help="filter (names=value) at the moment only for payload fields",
        action="append",
    )
    return parser


def main():
    options = mk_parser().parse_args()
    path = Path(options.path)
    filters = options.filter or []
    filters = [Filter(filter_str) for filter_str in filters]
    process_file_store(path, pl_type=options.type, filters=filters)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

# ############################################################################
# This is the code for THE pypeman command line program
# and all it's related commands
# after installing the module pypeman, all this functionality can be obtained
# by calling 'pypeman' from the command line
# ############################################################################


import asyncio
import logging
import os
import signal
import sys
import warnings

from functools import partial

import click

from DaemonLite import DaemonLite

# Keep this change of the python path
if True:
    CURRENT_DIR = os.getcwd()  # noqa: E402
    sys.path.insert(0, CURRENT_DIR)  # noqa: E402

import pypeman

from pypeman import channels
from pypeman import endpoints
from pypeman import nodes
from pypeman import remoteadmin
from pypeman.conf import settings
from pypeman.graph import load_project
from pypeman.graph import mk_graph
from pypeman.helpers.reloader import reloader_opt


CLI_CTX_SETTINGS = dict(help_option_names=['-h', '--help'])


async def sig_handler_coro(loop, signal, ctx):
    """
    asyncio code handling the reception of signals
    """
    # TODO: is this really the right way to stop pypeman.
    # TODO: should'nt we just stop the loop and let the main() function do the rest.
    logger = ctx["logger"]
    logger.debug("signal handler coro called for signal %s", signal)
    # TODO: Check probably this strange magic was intended to allow stopping
    #   file watchers or any other tasks with a sleep.
    #   Let's debug with a file watcher (currently using a sleep)
    #   and a channel with a sleep
    #
    # probably interrupting sleepers and implementing a grace period should not be done
    # here, but in "main()"
    #
    # logger.debug("wait for a grace periond")
    # grace_period = 3
    # await asyncio.sleep(grace_period)
    logger.debug("now stop the loop")
    loop.stop()

    # I really hope we can remove this code. and handle everything in main
    # for channel in channels.all_channels:
    #     logger.debug("stop channel %s", channel.name)
    #     await channel.stop()
    # # sleep for a grace period of 3 seconds. Might be able to remove lateron
    # logger.debug("waiting for %ds to let channels stop", grace_period)
    # logger.debug("loop stopped")


def sig_handler_func(loop, signal, ctx):
    loop.create_task(sig_handler_coro(loop, signal, ctx))


# TODO: might rename main and refactor into pypman.graph
# could call it perhaps. run_graph() or load_and_run_graph()
def main(debug_asyncio=False, profile=False, cli=False, remote_admin=False):

    if debug_asyncio:
        # set before asyncio.get_event_loop() call
        os.environ['PYTHONASYNCIODEBUG'] = "1"

    load_project()
    print('\nStarting...')

    logger = logging.getLogger(__name__)
    loop = asyncio.get_event_loop()

    from pypeman.plugin_mgr import manager as plugin_manager
    plugin_manager.set_loop(loop)

    ctx = dict(
        logger=logger,
        loop=loop,
        )

    signal_handler = partial(sig_handler_func, ctx=ctx)

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    signal.signal(signal.SIGHUP, signal_handler)

    if debug_asyncio:
        loop.slow_callback_duration = settings.DEBUG_PARAMS['slow_callback_duration']
        loop.set_debug(True)
        warnings.simplefilter('default')

    # Start plugins
    plugin_manager.start_plugins()

    # Start channels
    for chan in channels.all_channels:
        loop.run_until_complete(chan.start())

    # add signal handlers
    ctx = dict(  # a context dict to easily add more params if needed
        logger=logger,
        )

    for sig in (
            signal.SIGINT,  # CTRL-C
            signal.SIGTERM,  # Termination (kill w/o params)
            # SIGHUP,  # unused could use to reread / conf / reload nicely
            # SUGUSR1,  # could be used to print / dump debug info
            ):
        loop.add_signal_handler(sig, partial(sig_handler_func, loop, sig, ctx))

    # And endpoints
    for end in endpoints.all_endpoints:
        loop.run_until_complete(end.start())

    # At the moment mixing the asyncio and ipython
    # event loop is not working properly.
    # Thus ipython is executed in a separate thread
    if cli:
        from pypeman.helpers.cli import CLI
        namespace = dict(
            fake_signal_int=partial(sig_handler_func, signal.SIGINT, None),
            asyncio=asyncio,
            logger=logger,
            loop=loop,
            nodes=nodes,
            endpoints=endpoints,
            channels=channels,
            plugin_manager=plugin_manager,
            settings=settings,
            print_graph=print_graph,
            )
        ctx["cli"] = cli
        ctx.update(namespace)
        cli = CLI(namespace=namespace)
        cli.run_as_thread()

    # https://github.com/mhcomm/pypeman/issues/149 (convert rmt admin to plugin)
    if remote_admin:
        remote = remoteadmin.RemoteAdminServer(loop=loop, **settings.REMOTE_ADMIN_WEBSOCKET_CONFIG)
        loop.run_until_complete(remote.start())

        print("Remote admin websocket started at {host}:{port}".format(
            **settings.REMOTE_ADMIN_WEBSOCKET_CONFIG
        ))

        webadmin = remoteadmin.WebAdmin(loop=loop, **settings.REMOTE_ADMIN_WEB_CONFIG)

        loop.run_until_complete(webadmin.start())

        print("Web remote admin started at {host}:{port}".format(
            **settings.REMOTE_ADMIN_WEB_CONFIG
        ))

    print('Waiting for messages...')
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass

    logger.debug("loop was stopped.")
    logger.debug("stop pypeman graph as nicely as possible")
    # https://github.com/mhcomm/pypeman/issues/150 ( ensure clean shutdown)

    for end in endpoints.all_endpoints:
        logger.debug("stop endpoint %s", repr(end))
        loop.run_until_complete(end.stop())

    # Stop all channels
    for chan in channels.all_channels:
        logger.debug("stop channel %s", repr(chan))
        loop.run_until_complete(chan.stop())

    # Stop all plugins
    plugin_manager.stop_plugins()

    print("End started tasks...")
    pending = asyncio.all_tasks()
    logger.debug("%d pending tasks to wait for", len(pending))
    for task in pending:
        logger.debug("shall wait for task %s", repr(task))
    # TODO: This is where we might have to add a grace period.
    loop.run_until_complete(asyncio.gather(*pending))

    loop.close()


def mk_daemon(mainfunc=lambda: None, pidfile="pypeman.pid"):
    # TODO: might move to a separate module like e.g.  pypeman.helpers.daemon
    # might also look at following alternative modules:
    # - python-daemon
    # - daemonocle
    # - py daemoniker
    # Alternatively if we don't want other module dependencies we might just copy
    # the DaemonLite files into your source repository
    class DaemonizedApp(DaemonLite):
        def run(self):
            mainfunc()

    app = DaemonizedApp(pidfile)

    return app


@click.group()
@click.version_option(pypeman.__version__)
@click.help_option("--help", "-h")
def cli(version=False):
    """ Pypeman is a minimalistic but pragmatic ESB/ETL in python """


# some weird issue with new flake8 linters obliges us to add spaces before and after
# the '=' characters as soon as we add annotation strings
@cli.command(context_settings=CLI_CTX_SETTINGS)
@click.option(
    "--reload",
    is_flag=True,
    help="Make server autoreload (Dev only)",
    )
@click.option(
    "--debug-asyncio",
    is_flag=True,
    help="Enable asyncio debug",
    )
@click.option(
    "--cli",
    is_flag=True,
    help="enables an IPython CLI for debugging (not operational)",
    )
@click.option(
    "--remote-admin",
    is_flag=True,
    help="Enable remote admin server",
    )
@click.option(
    "--profile",
    is_flag=True,
    help="enables profiling / run stats (not operational)",
    )
@click.option(
    "--daemon/--no-daemon",
    default=True,
    help="started pypeman as daemon (default=True)",
    )
def start(reload, debug_asyncio, cli, remote_admin, profile, daemon):
    """ Start pypeman as daemon (or foreground process) """

    main_func = partial(
        main,
        debug_asyncio=debug_asyncio,
        cli=cli,
        profile=profile,
        remote_admin=remote_admin
    )
    start_func = partial(reloader_opt, main_func, reload, 2)

    daemon = (os.environ.get('PYPEMAN_NO_DAEMON') != "True") and daemon

    if reload:
        start_func()
    else:
        if daemon:
            daemon = mk_daemon(main_func)
            daemon.start()
        else:
            main_func()


@cli.command(context_settings=CLI_CTX_SETTINGS)
def stop():
    """ stops an already running pypeman instance """
    daemon = mk_daemon()
    daemon.stop()


@cli.command(context_settings=CLI_CTX_SETTINGS)
def printsettings():
    """
    print the project's settings

    """
    # TODO: This command could be enhanced further to be more simniliar to
    #       django_extensions' print_settings
    load_project()
    svars = sorted([val for val in dir(settings) if "A" <= val[0] <= "Z"])
    for name in svars:
        print(name, "=", repr(getattr(settings, name)))


def print_graph(dot=False):
    """ prints the project graph """
    for line in mk_graph(dot):
        print(line)


@cli.command(context_settings=CLI_CTX_SETTINGS)
@click.option(
    "--dot", is_flag=True,
    help="Make dot compatible output (Can be viewed with http://ushiroad.com/jsviz/)",
    )
def graph(dot):
    """ Show channel graph"""

    load_project()
    print_graph(dot)


@cli.command(context_settings=CLI_CTX_SETTINGS)
def pyshell():
    """ Start ipython shell to send command to remote instance """
    client = remoteadmin.RemoteAdminClient(
        url='ws://%s:%s' % (
            settings.REMOTE_ADMIN_WEBSOCKET_CONFIG['host'],
            settings.REMOTE_ADMIN_WEBSOCKET_CONFIG['port']))
    client.init()

    from IPython import embed
    embed()


@cli.command(context_settings=CLI_CTX_SETTINGS)
def shell():
    """ Start a custom shell to administrate remote pypeman instance """
    settings.init_settings()
    try:
        remoteadmin.PypemanShell(url='ws://%s:%s' % (settings.REMOTE_ADMIN_WEBSOCKET_CONFIG['host'],
                                 settings.REMOTE_ADMIN_WEBSOCKET_CONFIG['port'])).cmdloop()
    except KeyboardInterrupt:
        print('\nQuitting...')


@cli.command(context_settings=CLI_CTX_SETTINGS)
@click.argument("dirname")
def startproject(dirname):
    """ Creates a pypeman project from scratch

    DIRNAME: name of dir to install project to
    """
    from pypeman.pjt_templates import new_project
    new_project(dirname)


@cli.command(context_settings=CLI_CTX_SETTINGS)
def debug():
    """ Used for development purpose """
    pass


@cli.command(context_settings={
    "ignore_unknown_options": True, **CLI_CTX_SETTINGS})
@click.option(
    "--module", "-m",
    default="tests",
    help="the module parameter for unittest.main() (default: tests)",
    )
@click.argument("args", nargs=-1)
@click.pass_context
def test(ctx, module, args):
    """ Launch project's tests with unittest.main().
        All tests from one module (default 'tests') will be executed.

        MODULE: name of module (and its submodules) to test default=(tests)

        ARGS:   further args for unittest.main().
                To get more help type: pypeman test -- -h
    """
    args = [arg for arg in args if arg != "--"]
    from unittest import main

    load_project()

    # unittest.main should call sys.exit(). So no need to return status code
    # to caller level.
    # we could pass exit=False if we wanted to not call exit and do a custom
    # treatment
    main(module=module, argv=['pypeman test --'] + list(args))


@cli.command(context_settings={
    "ignore_unknown_options": True, **CLI_CTX_SETTINGS})
@click.argument("args", nargs=-1)
@click.pass_context
def pytest(ctx, args):
    """ start tests with pytest.
        Params can be passed through with -- [args...].
        Pytest help with:
        pypeman pytest -- -h
    """
    import pytest

    load_project()
    if not args:
        args = ["tests.py"]
    rslt = pytest.main(list(args))
    sys.exit(rslt)


if __name__ == "__main__":
    cli()



from setuptools import setup
from importlib import import_module

VERSION = import_module("pypeman").__version__
URL = import_module("pypeman").__url__

with open("./README.rst") as desc_file:
    long_description = desc_file.read()

setup(
    name="pypeman",
    version=VERSION,
    description="Minimalistic but pragmatic ESB / ETL / EAI in Python",
    long_description=long_description,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Healthcare Industry",
        "Operating System :: POSIX :: Linux",
        "Topic :: Database",
        "Topic :: Internet",
        "Topic :: Internet :: File Transfer Protocol (FTP)",
        "Programming Language :: Python",
    ],
    keywords="esb etl eai pipeline data processing asyncio http ftp hl7",
    url=URL,

    author="Jeremie Pardou",
    author_email="jeremie@jeremiez.net",

    license="Apache Software License",
    packages=[
        "pypeman",
        "pypeman.contrib",
        "pypeman.helpers",
        "pypeman.plugins",
        "pypeman.plugins.remoteadmin",
        "pypeman.tests",
        "pypeman.tests.test_app",
        "pypeman.tests.test_app_testing",
        "pypeman.tests.settings",
        "pypeman.tools",
    ],

    package_data={
        "pypeman.tests": ["data/*"],
    },
    entry_points={
        "console_scripts": [
            "pypeman = pypeman.commands:cli",
            "pypeman_tool = pypeman.tool_commands:main",
        ]
    },

    test_suite="nose.collector",
    install_requires=[
        "click",
        "daemonlite",
        "python-dateutil",
        "websockets",
        "aiohttp",
        "jsonrpcclient",
        "jsonrpcserver<5",  # cannot import name 'Protocol' from 'typing' : fixed in py3.8
        "sqlitedict",
        # TODO Remove next dependency as soon as jsonrpcclient is removed
        "requests"
    ],
    extras_require={
        "hl7": ["hl7"],
        "xml": ["xmltodict"],
        "time": ["aiocron"],
        "all": ["hl7", "xmltodict", "aiocron"]
    },
    setup_requires=["pytest-runner"],
    tests_require=[
        "pytest", "pytest-cov", "pytest-asyncio"
        ],
    include_package_data=True,
)



import asyncio
import contextvars
import logging
import re
import traceback
import types
import uuid
import warnings

from pathlib import Path

from pypeman import message, msgstore, events
from pypeman.errors import PypemanConfigError
from pypeman.helpers.itertools import flatten
from pypeman.helpers.sleeper import Sleeper


logger = logging.getLogger(__name__)

# List all channel registered
all_channels = []

_channel_names = set()

MSG_CTXVAR = contextvars.ContextVar("msg")


class Dropped(Exception):
    """ Used to stop process as message is processed. Default success should be returned.
    """


class Rejected(Exception):
    """ Used to tell caller the message is invalid with a error return.
    """


class ChannelStopped(Exception):
    """ The channel is stopped and can't process message.
    """


class BaseChannel:
    """
    Base channel are generic channels.
    If you want to create new channel, inherit from the base class and call
    ``self.handle(msg)`` method with generated message.

    :param name: Channel name is mandatory and must be unique through the whole project.
        Name gives a way to get channel in test mode.

    :param parent_channel: Used with sub channels. Don't specify yourself.

    :param loop: To specify a custom event loop.

    :param message_store_factory:     You can specify a message store (see below) at channel
        initialisation if you want to save all processed message. Use
        `message_store_factory` argument with  an instance of wanted message store factory.

    :param wait_subchans: Boolean, if set to True, channels will wait for suchannel ends for sending,
    speed response and process a new message
    """
    STARTING, WAITING, PROCESSING, STOPPING, STOPPED = range(5)
    STATE_NAMES = ['STARTING', 'WAITING', 'PROCESSING', 'STOPPING', 'STOPPED']

    def __init__(self, name=None, parent_channel=None, loop=None, message_store_factory=None,
                 wait_subchans=False, verbose_name=None):

        self.uuid = uuid.uuid4()

        all_channels.append(self)
        self._nodes = []
        self._node_map = {}
        self._status = BaseChannel.STOPPED
        self.processed_msgs = 0
        self.interruptable_sleeper = Sleeper(loop)  # for interruptable sleeps
        self.join_nodes = None
        self.fail_nodes = None
        self.drop_nodes = None
        self.reject_nodes = None
        self.final_nodes = None
        self.wait_subchans = wait_subchans
        self.raise_dropped = False

        if name:
            self.name = name
        else:
            warnings.warn(
                "Channels without names are deprecated and will be removed in version 0.5.2",
                DeprecationWarning)
            self.name = self.__class__.__name__ + "_" + str(len(all_channels))

        self.parent = None
        if parent_channel:
            # Use dot name hierarchy
            self.name = ".".join([parent_channel.name, self.name])

            self.parent = parent_channel

            # TODO parent channels usefull ?
            self.parent_uids = [parent_channel.uuid]
            self.parent_names = [parent_channel.name]
            if parent_channel.parent_uids:
                self.parent_uids.append(parent_channel.parent_uids)
                self.parent_names += parent_channel.parent_names
        else:
            self.parent_uids = None

        if self.name in _channel_names:
            raise NameError(
                "Duplicate channel name %r . "
                "Channel names must be unique !" % self.name
            )
        if verbose_name:
            self.verbose_name = verbose_name
        else:
            self.verbose_name = self.name

        _channel_names.add(self.name)

        if loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = loop

        self.logger = logging.getLogger('pypeman.channels.%s' % self.name)

        self.next_node = None

        self.message_store_factory = message_store_factory or msgstore.NullMessageStoreFactory()

        self.message_store = self.message_store_factory.get_store(self.name)

        self._first_start = True

        # Used to avoid multiple messages processing at same time
        # Lock use `asyncio.get_running_loop()` that only be called from coroutines or callbacks
        # So now, lock is instanciated at start
        self.lock = None

        self.sub_chan_tasks = []
        self.sub_chan_endnodes = []

    def _reset_sub_chan_endnodes(self, fut):
        """
        Remove all subchan callbacks from the list
        called by the done_callback of subchan future callbacks
        """
        self.sub_chan_endnodes = []
        fut.result()

    @classmethod
    def status_id_to_str(cls, state_id):
        return cls.STATE_NAMES[state_id]

    @classmethod
    def status_str_to_id(cls, state):
        return cls.STATE_NAMES.index(state)

    @property
    def status(self):
        """ Getter for status """
        return self._status

    @status.setter
    def status(self, value):
        old_state = self._status
        self._status = value
        # Launch change state event
        asyncio.create_task(events.channel_change_state.fire(
            channel=self, old_state=old_state, new_state=value))

    def is_stopped(self):
        """
        :return: True if channel is in stopped or stopping state.
        """
        return self.status in (BaseChannel.STOPPING, BaseChannel.STOPPED,)

    async def start(self):
        """
        Start the channel. Called before starting processus. Can be overloaded to specify specific
        start procedure.
        """
        self.logger.debug("Channel %s starting ...", str(self))
        self.lock = asyncio.Lock()
        self.status = BaseChannel.STARTING
        if self._first_start:
            self.init_node_graph()
            self._first_start = False
        await self.message_store.start()
        self.status = BaseChannel.WAITING
        self.logger.info("Channel %s started", str(self))

    def init_node_graph(self):
        if self._nodes:
            previous_node = self._nodes[0]

            for node in self._nodes[1:]:
                previous_node.next_node = node
                previous_node = node

    async def stop(self):
        """
        Stop the channel.
        Called when
        - pypeman shuts down.
        - a channel is stopped (e.g. via the admin interface)
        """
        self.logger.debug("Channel %s stopping ...", str(self))
        self.status = BaseChannel.STOPPING
        # Verify that all messages are processed
        async with self.lock:
            self.status = BaseChannel.STOPPED
        # stop all pending sleeps
        await self.interruptable_sleeper.cancel_all()
        self.logger.info("Channel %s stopped", str(self))

    def _reset_test(self):
        """ Enable test mode and reset node data.
        """
        self.raise_dropped = True
        for node in self._nodes:
            node._reset_test()
        if self.join_nodes:
            for node in self.join_nodes:
                node._reset_test()
        if self.drop_nodes:
            for node in self.drop_nodes:
                node._reset_test()
        if self.reject_nodes:
            for node in self.reject_nodes:
                node._reset_test()
        if self.final_nodes:
            for node in self.final_nodes:
                node._reset_test()
        if self.fail_nodes:
            for node in self.fail_nodes:
                node._reset_test()

    def add(self, *args):
        """
        Add specified nodes to channel (Shortcut for append).

        :param args: Nodes to add.
        """
        return self.append(*args)

    def _register_node(self, node):
        """
        Node registering to search.

        :param node: node to register.
        """
        self._node_map[node.name] = node
        if self.parent:
            self.parent._register_node(node)

    def get_node(self, name):
        """
        Return node with name in argument. Mainly used in tests.

        :param name: The searched node name.

        :return: Instance of Node or None if none found.
        """

        return self._node_map.get(name)

    def append(self, *args):
        """
        Append specified nodes to channel.

        :param args: Nodes to add. can be a list, a tupe or a nested
                list of tuples or lists if nodes.
                nodes with value None are ignored
        """
        args = [node for node in flatten(args) if node]
        for node in args:
            node.channel = self
            self._nodes.append(node)
            self._register_node(node)

        return self

    def fork(self, name=None, message_store_factory=None):
        """
        Create a new channel that processes a copy of the message
        at this point.
        Subchannels are executed in parallel of main process.

        :return: The forked channel
        """

        s = SubChannel(
            name=name, parent_channel=self,
            message_store_factory=message_store_factory, loop=self.loop,
            wait_subchans=self.wait_subchans)
        self._nodes.append(s)
        return s

    def when(self, condition, name=None, message_store_factory=None):
        """
        New channel bifurcation that is executed only if condition is True. This channel
        replace further current channel processing.

        :param condition: Can be a value or a function that takes a message argument.

        :return: The conditional path channel.
        """

        s = ConditionSubChannel(
            condition=condition, name=name, parent_channel=self,
            message_store_factory=message_store_factory, loop=self.loop,
            wait_subchans=self.wait_subchans)
        self._nodes.append(s)
        return s

    def case(self, *conditions, names=None, message_store_factory=None):
        """
        Case between multiple conditions. For each condition specified, a
        channel is returned by this method in same order as condition are
        given.
        When processing a message, conditions are evaluated successively and
        the first returning true triggers the corresponding channel processing
        the message.
        When channel processing is finished, the next node is called.

        :param conditions: Multiple conditions, one for each returned channel.
            Should be a boolean or a function that takes a ``msg`` argument
            and should return a boolean.

        :param message_store_factory: Allows you to specify a message store
            factory for all channel of this `case`.

        :return: one channel by condition parameter.
        """
        if names is None:
            names = [None] * len(conditions)

        c = Case(*conditions, names=names, parent_channel=self,
                 message_store_factory=message_store_factory, loop=self.loop,
                 wait_subchans=self.wait_subchans)
        self._nodes.append(c)
        return [chan for cond, chan in c.cases]

    def handle_and_wait(self, msg):
        """ Handle a message synchronously. Mainly used for testing purpose.

        :param msg: Message to process
        :return: Processed message.
        """
        return self.loop.run_until_complete(self.handle(msg))

    async def handle(self, msg):
        """ Overload this method only if you know what you are doing but
        call it from child class to add behaviour.

        :param msg: To be processed msg.
        :return: Processed message
        """

        # Store message before any processing
        # TODO If store fails, do we stop processing ?
        # TODO Do we store message even if channel is stopped ?
        msg_store_id = await self.message_store.store(msg)

        if self.status in [BaseChannel.STOPPED, BaseChannel.STOPPING]:
            raise ChannelStopped("Channel is stopped so you can't send message.")
        self.logger.info("chan %s handle %s", str(self), str(msg))
        has_callback = hasattr(self, "_callback")
        setattr(msg, "chan_rslt", None)
        setattr(msg, "chan_exc", None)
        setattr(msg, "chan_exc_traceback", None)
        # Only one message processing at time
        async with self.lock:
            self.status = BaseChannel.PROCESSING
            try:
                result = await self.subhandle(msg.copy())
                await self.message_store.change_message_state(msg_store_id, message.Message.PROCESSED)
                msg.chan_rslt = result
                if self.join_nodes and not has_callback:
                    await self.join_nodes[0].handle(result.copy())
                return result
            except Dropped as exc:
                self.logger.info("%s DROP msg %s", str(self), str(msg))
                msg.chan_exc = exc
                msg.chan_exc_traceback = traceback.format_exc()
                await self.message_store.change_message_state(msg_store_id, message.Message.PROCESSED)
                if self.drop_nodes and not has_callback:
                    await self.drop_nodes[0].handle(msg.copy())
                if self.raise_dropped:
                    raise
                return msg
            except Rejected as exc:
                self.logger.info("%s REJECT msg %s", str(self), str(msg))
                msg.chan_exc = exc
                msg.chan_exc_traceback = traceback.format_exc()
                await self.message_store.change_message_state(msg_store_id, message.Message.REJECTED)
                await self.message_store.add_message_meta_infos(msg_store_id, "err_msg", str(exc))
                if self.reject_nodes and not has_callback:
                    await self.reject_nodes[0].handle(msg.copy())
                raise
            except Exception as exc:
                msg.chan_exc = exc
                msg.chan_exc_traceback = traceback.format_exc()
                self.logger.error('Error while processing message %s (chan %s)', str(msg), str(self))
                await self.message_store.change_message_state(msg_store_id, message.Message.ERROR)
                await self.message_store.add_message_meta_infos(msg_store_id, "err_msg", str(exc))
                if self.fail_nodes and not has_callback:
                    await self.fail_nodes[0].handle(msg.copy())
                raise
            finally:
                self.status = BaseChannel.WAITING
                self.processed_msgs += 1
                if self.final_nodes and not has_callback:
                    await self.final_nodes[0].handle(msg.copy())
                try:
                    if self.sub_chan_tasks:
                        # Launch sub chans handle()
                        subchantasks = asyncio.gather(*self.sub_chan_tasks)
                        if self.wait_subchans:
                            await subchantasks
                finally:
                    if self.sub_chan_endnodes:
                        # Launch and wait for sub chans callbacks
                        subchan_endnodes_fut = asyncio.gather(*self.sub_chan_endnodes)
                        subchan_endnodes_fut.add_done_callback(self._reset_sub_chan_endnodes)
                        if self.wait_subchans:
                            await subchan_endnodes_fut
                    self.logger.info("%s end handle %s", str(self), str(msg))

    async def subhandle(self, msg):
        """ Overload this method only if you know what you are doing. Called by ``handle`` method.

        :param msg: To be processed msg.

        :return: Processed message
        """

        result = await self.process(msg)

        if self.next_node:
            if isinstance(result, types.GeneratorType):
                gene = result
                result = msg  # Necessary if all nodes result are dropped
                for res in gene:
                    try:
                        result = await self.next_node.handle(res)
                    except Dropped:
                        pass
                    # TODO Here result is last value returned. Is it a good idea ?
            else:
                result = await self.next_node.handle(result)

        return result

    async def process(self, msg):
        """ Overload this method only if you know what you are doing. Called by ``subhandle`` method.

        :param msg: To be processed msg.

        :return: Processed message
        """

        if self._nodes:
            res = await self._nodes[0].handle(msg)
            return res
        else:
            return msg

    async def replay(self, msg_id):
        """
        This method allows you to replay a message from channel `message_store`.

        :param msg_id: Message id to replay.

        :return: The result of the processing.
        """
        self.logger.info("try to replay %s", str(msg_id))
        msg_dict = await self.message_store.get(msg_id)
        new_message = msg_dict['message'].renew()
        result = await self.handle(new_message)
        return result

    def to_dict(self):
        return {
            'name': self.name,
            'verbose_name': self.verbose_name,
            'status': BaseChannel.status_id_to_str(self.status),
            'has_message_store': not isinstance(self.message_store, msgstore.NullMessageStore),
            'processed': self.processed_msgs,
        }

    def subchannels(self):
        res = []

        for node in self._nodes:
            if isinstance(node, SubChannel) or isinstance(node, ConditionSubChannel):
                chan_dict = node.to_dict()
                chan_dict['subchannels'] = node.subchannels()
                res.append(chan_dict)
            elif isinstance(node, Case):
                for cond, channel in node.cases:
                    chan_dict = channel.to_dict()
                    chan_dict['subchannels'] = channel.subchannels()
                    res.append(chan_dict)
        return res

    def graph(self, prefix='', dot=False):
        """
        Generate a text graph for this channel.
        """
        # TODO: ask to klaus how to implement and show endnodes in the graph
        for node in self._nodes:
            if isinstance(node, SubChannel):
                yield prefix + '|—\\ (%s)' % node.name
                for entry in node.graph(prefix='|  ' + prefix):
                    yield entry
            elif isinstance(node, ConditionSubChannel):
                yield prefix + '|?\\ (%s)' % node.name
                for entry in node.graph(prefix='|  ' + prefix):
                    yield entry
                yield prefix + '|  -> Out'
            elif isinstance(node, Case):
                for i, c in enumerate(node.cases):
                    yield prefix + '|c%s\\' % i
                    for entry in c[1].graph(prefix='|  ' + prefix):
                        yield entry
                    yield prefix + '|<--'
            else:
                yield prefix + '|-' + node.name

    def graph_dot(self, end=''):
        """
        Generate a compatible dot graph for this channel.
        """
        after = []
        cases = None

        yield '#---'

        previous = self.name

        if end == '':
            end = self.name

        for node in self._nodes:
            if isinstance(node, SubChannel):
                yield '"%s"->"%s";' % (previous, node.name)
                after.append((None, node))

            elif isinstance(node, ConditionSubChannel):
                yield '"%s"->"%s" [style=dotted];' % (previous, node.name)
                after.append((end, node))

            elif isinstance(node, Case):
                cases = [c[1] for c in node.cases]

            else:
                if cases:
                    for c in cases:
                        yield '"%s"->"%s" [style=dotted];' % (previous, c.name)
                        after.append((node.name, c))
                    cases = None

                else:
                    yield '"%s"->"%s";' % (previous, node.name)

                previous = node.name

        if end:
            yield '"%s"->"%s";' % (previous, end)

        for end, sub in after:
            for entry in sub.graph_dot(end=end):
                yield entry

    def _init_end_nodes(self, *end_nodes):
        """
        join/fail/drop/reject/final nodes are nodes that are launched
        after channel subhandle
        This func permits to chain them without adding them at the end of self._nodes
        """
        end_nodes = [node for node in flatten(end_nodes) if node]
        for node in end_nodes:
            node.channel = self

        if len(end_nodes) > 1:
            previous_node = end_nodes[0]

            for node in end_nodes[1:]:
                previous_node.next_node = node
                previous_node = node
        return end_nodes

    def add_join_nodes(self, *end_nodes):
        """
        Add nodes that will be launched only after a successful channel process
        The first node take the result message of the channel as input

        CAUTION: TODO: BUG if an exception occurs during processing of join_nodes there's
        a different comportment if it's a BaseChannel or a SubChannel (due to implementation)
            - BaseChannel: if there's fail/drop/reject_nodes, they will be launched
            - SubChannel: no other endnodes will be launched (except final_nodes)
        """
        if self.join_nodes:
            raise PypemanConfigError(f"join_nodes already existing for channel {self.name}")
        self.join_nodes = self._init_end_nodes(*end_nodes)

    def add_fail_nodes(self, *end_nodes):
        """
        Add nodes that will be launched only after a channel process that raises an Exception
        (except Dropped and Rejected)
        The first node take the entry message of the channel as input
        """
        if self.fail_nodes:
            raise PypemanConfigError(f"fail_nodes already existing for channel {self.name}")
        self.fail_nodes = self._init_end_nodes(*end_nodes)

    def add_drop_nodes(self, *end_nodes):
        """
        Add nodes that will be launched only after a channel process that raises a Dropped
        The first node take the entry message of the channel as input
        """
        if self.drop_nodes:
            raise PypemanConfigError(f"drop_nodes already existing for channel {self.name}")
        self.drop_nodes = self._init_end_nodes(*end_nodes)

    def add_reject_nodes(self, *end_nodes):
        """
        Add nodes that will be launched only after a channel process that raises a Rejected
        The first node take the entry message of the channel as input
        """
        if self.reject_nodes:
            raise PypemanConfigError(f"reject_nodes already existing for channel {self.name}")
        self.reject_nodes = self._init_end_nodes(*end_nodes)

    def add_final_nodes(self, *end_nodes):
        """
        Add nodes that will be launched all the time after a channel process
        The first node take the entry message of the channel as input (TODO: ask if it's ok)
        """
        if self.final_nodes:
            raise PypemanConfigError(f"final_nodes already existing for channel {self.name}")
        self.final_nodes = self._init_end_nodes(*end_nodes)

    # def _callback(self, future):
    #     """
    #     Function called by subchannel future done_callback
    #     If this function is implemented join/fail/reject/drop/final endnodes are not instanciated
    #     in handle but in this func so take care to not implement it if you are not sure what
    #     you're doing
    #     TODO: Maybe add a warning if _callback and not isinstance(SubChannel)  ??
    #     """
    #     pass

    def __str__(self):
        return "<chan: %s>" % self.name

    def __repr__(self):
        typ = type(self)
        return "<%s.%s:(\"%s\")>" % (typ.__module__, typ.__name__, self.name)


def reset_pypeman_channels():
    """
    clears book keeping of all channels

    Can be useful for unit testing.
    """
    logger.info("clearing all_channels and _channel-names.")
    all_channels.clear()
    _channel_names.clear()


class SubChannel(BaseChannel):
    """ Subchannel used for forking channel processing. """

    def _callback(self, fut):
        """
        """
        ctx = contextvars.copy_context()
        entrymsg = ctx.get(MSG_CTXVAR)
        setattr(entrymsg, "chan_rslt", None)
        setattr(entrymsg, "chan_exc", None)
        setattr(entrymsg, "chan_exc_traceback", None)
        endnodes_tasks = []
        try:
            result = fut.result()
            entrymsg.chan_rslt = result
            if self.join_nodes:
                endnode_task = asyncio.create_task(self.join_nodes[0].handle(result.copy()))
                endnodes_tasks.append(endnode_task)
            logger.info(
                "Subchannel %s end process message %s, rslt is msg %s",
                str(self), str(entrymsg), str(result))
        except Dropped as exc:
            entrymsg.chan_exc = exc
            entrymsg.chan_exc_traceback = traceback.format_exc()
            if self.drop_nodes:
                endnode_task = asyncio.create_task(self.drop_nodes[0].handle(entrymsg.copy()))
                endnodes_tasks.append(endnode_task)
            self.logger.info("Subchannel %s. Msg %s was dropped", str(self), str(entrymsg))
        except Rejected as exc:
            entrymsg.chan_exc = exc
            entrymsg.chan_exc_traceback = traceback.format_exc()
            if self.reject_nodes:
                endnode_task = asyncio.create_task(self.reject_nodes[0].handle(entrymsg.copy()))
                endnodes_tasks.append(endnode_task)
            self.logger.info("Subchannel %s. Msg %s was Rejected", str(self), str(entrymsg))
            raise
        except Exception as exc:
            entrymsg.chan_exc = exc
            entrymsg.chan_exc_traceback = traceback.format_exc()
            if self.fail_nodes:
                endnode_task = asyncio.create_task(self.fail_nodes[0].handle(entrymsg.copy()))
                endnodes_tasks.append(endnode_task)
            self.logger.exception(
                "Error while processing msg %s in subchannel %s", str(entrymsg), str(self))
            raise
        finally:
            if self.final_nodes:
                endnode_task = asyncio.create_task(self.final_nodes[0].handle(entrymsg.copy()))
                endnodes_tasks.append(endnode_task)
            self.parent.sub_chan_endnodes.extend(endnodes_tasks)
            self.parent.sub_chan_tasks.remove(fut)
            self.logger.info(
                "subchan %s end process msg %s", str(self), str(entrymsg))

    async def process(self, msg):
        if self._nodes:
            msgctxvartoken = MSG_CTXVAR.set(msg.copy())
            ctx = contextvars.copy_context()
            fut = asyncio.create_task(self._nodes[0].handle(msg.copy()))
            fut.add_done_callback(self._callback, context=ctx)
            self.parent.sub_chan_tasks.append(fut)
            MSG_CTXVAR.reset(msgctxvartoken)
        return msg


class ConditionSubChannel(BaseChannel):
    """
    ConditionSubchannel used for make alternative path. This processing replace
    all further channel processing.
    """

    def __init__(self, condition=lambda x: True, **kwargs):
        super().__init__(**kwargs)
        self.condition = condition

    def test_condition(self, msg):
        if callable(self.condition):
            return self.condition(msg)
        else:
            return self.condition

    async def subhandle(self, msg):
        result = await self.process(msg)
        return result

    async def handle(self, msg):
        if self.test_condition(msg):
            result = await super().handle(msg)
        else:
            if self.next_node:
                result = await self.next_node.handle(msg)
            else:
                result = msg

        return result


class Case():
    """ Case node internally used for `.case()` BaseChannel method. Don't use it.
    """
    def __init__(self, *args, names=None, parent_channel=None, message_store_factory=None, loop=None,
                 wait_subchans=False):
        self.next_node = None
        self.cases = []

        if names is None:
            names = []

        self.loop = loop or asyncio.get_event_loop()

        if message_store_factory is None:
            message_store_factory = msgstore.NullMessageStoreFactory()

        for cond, name in zip(args, names):
            b = BaseChannel(name=name, parent_channel=parent_channel,
                            message_store_factory=message_store_factory,
                            loop=self.loop, wait_subchans=wait_subchans)
            self.cases.append((cond, b))

    def _reset_test(self):
        for c in self.cases:
            c[1]._reset_test()

    def test_condition(self, condition, msg):
        if callable(condition):
            return condition(msg)
        else:
            return condition

    async def handle(self, msg):
        result = msg
        for cond, channel in self.cases:
            if self.test_condition(cond, msg):
                result = await channel.handle(msg)
                break

        if self.next_node:
            result = await self.next_node.handle(result)

        return result


class MergeChannel(BaseChannel):
    """
    This class permits to have multiple channel classes (watchers/listeners/..)
    as inputs to a single channel
    TODO: CAUTION: Not sure that input channels works with end nodes + they don't
        appears in graph

    TODO: check why we need
        if channel.loop != self.loop:
            channel.loop = self.loop
    There's a bug with event loop (maybe only on tests) but weird
    (traceback: RuntimeError: Task <Task pending name='Task-126' coro=<MergeChannel.start()
    running at
    /home/quentin/projects/interop2/app/vendor/common/custom_channels.py:51>
    cb=[_run_until_complete_cb()
    at /home/quentin/.pyenv/versions/3.10.10/lib/python3.10/asyncio/base_events.py:184]>
    got Future
    <Future pending cb=[_chain_future.<locals>._call_check_cancel() at
    /home/quentin/.pyenv/versions/3.10.10/lib/python3.10/asyncio/futures.py:385]>
    attached to a different loop
    )

    Args:
        channels (list of channels): List of channels to use as inputs
    """

    def __init__(self, *args, chans, **kwargs):
        super().__init__(*args, **kwargs)
        self.channels = chans
        for channel in self.channels:
            all_channels.remove(channel)
            channel.add = None
            channel.handle = self.handle
            channel.handle_and_wait = self.handle_and_wait
            if channel.loop != self.loop:
                channel.loop = self.loop

    async def start(self):
        for channel in self.channels:
            if channel._nodes:
                raise AttributeError(f"A merged channel cannot have nodes (chan {channel.name})")
            if channel.loop != self.loop:
                channel.loop = self.loop
            await channel.start()
        await super().start()

    async def stop(self):
        for channel in self.channels:
            if channel.loop != self.loop:
                channel.loop = self.loop
            await channel.stop()
        await super().stop()


class FileWatcherChannel(BaseChannel):
    """
    Watch for file change or creation. File content becomes message payload.
    ``filepath`` is in message meta.

    If the regex is for an acknowledgement file (.ok for example) you can convert it to the
    real filepath via the real_extensions init arg. The returned msg payload will be the
    content of the real file and not the acknowledgement file. Idem for meta
    """

    NEW, UNCHANGED, MODIFIED, DELETED = range(4)

    def __init__(self, *args, basedir='', regex='.*', interval=1, binary_file=False, path='',
                 real_extensions=None, **kwargs):
        super().__init__(*args, **kwargs)
        if path:
            self.basedir = path
            warnings.warn("path deprecated, use basedir instead", DeprecationWarning)
        if basedir:
            self.basedir = basedir
        self.basedir = Path(self.basedir)
        self.regex = regex
        self.interval = interval
        self.dirflag = self.basedir.is_dir()
        self.data = {}
        self.re = re.compile(self.regex)
        self.binary_file = binary_file
        self.real_extensions = real_extensions  # list of extensions for exemple: [".csv", ".CSV"]
        # Set mtime for all existing matching files
        if self.basedir.exists():
            for filepath in self.basedir.iterdir():
                if self.re.match(filepath.name):
                    mtime = filepath.stat().st_mtime
                    self.data[filepath.name] = mtime
        else:
            self.logger.warning("Path doesn't exists: %r", self.basedir)

    async def start(self):
        await super().start()
        asyncio.create_task(self.watch_for_file())

    def file_status(self, filename):
        if filename in self.data:
            old_mtime = self.data[filename]
            filepath = self.basedir / filename
            new_mtime = filepath.stat().st_mtime
            if new_mtime == old_mtime:
                return FileWatcherChannel.UNCHANGED
            elif new_mtime > old_mtime:
                return FileWatcherChannel.MODIFIED
        else:
            return FileWatcherChannel.NEW

    def _handle_callback(self, future):
        try:
            future.result()
        except Dropped:
            pass

    async def watch_for_file(self):
        # self.logger.debug("Will sleep")
        await self.interruptable_sleeper.sleep(self.interval)
        # await asyncio.sleep(self.interval, loop=self.loop)
        # self.logger.debug("sleep done")
        try:
            if self.basedir.exists():
                listfile = self.basedir.iterdir()

                for filepath in sorted(listfile):
                    filename = filepath.name
                    if self.re.match(filename):
                        status = self.file_status(filename)

                        if status in [FileWatcherChannel.MODIFIED, FileWatcherChannel.NEW]:
                            self.data[filename] = filepath.stat().st_mtime
                            if self.real_extensions:
                                for extension in self.real_extensions:
                                    real_fpath = filepath.with_suffix(extension)
                                    if real_fpath.exists():
                                        filepath = real_fpath
                                        filename = real_fpath.name
                                        break
                                else:
                                    # If no related files
                                    # TODO : ask if raise exc or not
                                    logger.error(
                                        "No %r related file to %s",
                                        self.real_extensions, str(filepath))
                                    continue

                            # Read file and make message
                            if self.binary_file:
                                mode = "rb"
                            else:
                                mode = "r"

                            msg = message.Message()

                            with filepath.open(mode) as f:
                                msg.payload = f.read()
                            msg.meta['filename'] = filename
                            msg.meta['filepath'] = str(filepath)
                            fut = asyncio.create_task(self.handle(msg))
                            fut.add_done_callback(self._handle_callback)

        except Exception:  # TODO: might explicitely silence some special cases.
            self.logger.exception("filewatcher problem")
        finally:
            if self.status not in (BaseChannel.STOPPING, BaseChannel.STOPPED,):
                asyncio.create_task(self.watch_for_file())
            else:
                logger.warning("Won't watch anymore")


from pypeman.helpers import lazyload  # noqa: E402

wrap = lazyload.Wrapper(__name__)

wrap.add_lazy('pypeman.contrib.hl7', 'MLLPChannel', ['hl7'])
wrap.add_lazy('pypeman.contrib.http', 'HttpChannel', ['aiohttp'])
wrap.add_lazy('pypeman.contrib.time', 'CronChannel', ['aiocron'])
wrap.add_lazy('pypeman.contrib.ftp', 'FTPWatcherChannel', [])



"""
Global graph related pypeman functions.

On the long run some commands of pypeman.commands might be candidates
to be moved into this module
"""


import importlib
import sys
import time
import traceback

from pypeman.conf import settings
from pypeman import channels
from pypeman.errors import PypemanError


def load_project():
    settings.init_settings()
    from pypeman.plugin_mgr import manager as plugin_manager
    plugin_manager.import_plugins()
    project_module = settings.PROJECT_MODULE
    try:
        importlib.import_module(project_module)
    except ImportError as exc:
        msg = str(exc)
        if 'No module' not in msg:
            print("IMPORT ERROR %s" % project_module)
            raise
        if project_module not in msg:
            print("IMPORT ERROR %s" % project_module)
            raise
        print("Missing '%s' module !" % project_module)
        sys.exit(-1)
    except Exception:
        traceback.print_exc()
        raise
    plugin_manager.init_plugins()
    plugin_manager.ready_plugins()


def wait_for_loop(tmax=5.0):
    """
    wait until the loop variable of a pypeman graph
    has been initialized

    can be used from any thread that's not the main thread
    """
    # TODO: might factor out this function to a helper module
    loop = None
    steps = int(tmax / 0.1)
    for i in range(steps, -1, -1):
        try:
            channel = channels.all_channels[0]
            # print("channel =", channel)
            loop = channel.loop
            break
        except Exception:
            if i == 0:
                raise PypemanError("couldn't obtain graph's loop")
        time.sleep(0.1)
    return loop


def mk_ascii_graph(title=None):
    """ Show pypeman graph as ascii output.
        Better reuse for debugging or new code
    """
    if title:
        yield title
    for channel in channels.all_channels:
        if not channel.parent:
            yield channel.__class__.__name__
            for entry in channel.graph():
                yield entry
            yield "|-> out"
            yield ""


def mk_graph(dot=False):
    if dot:
        yield "digraph testgraph{"

        # Handle channel node shape
        for channel in channels.all_channels:
            yield '{node[shape=box]; "%s"; }' % channel.name

        # Draw each graph
        for channel in channels.all_channels:
            if not channel.parent:
                for line in channel.graph_dot():
                    yield line

        yield "}"
    else:
        for line in mk_ascii_graph():
            yield line



#!/usr/bin/env python

# ############################################################################
# Name       : cli.py
"""
  Summary    :  simple ipython debug CLI with basic fallback

__author__    = "Klaus Foerster"
"""
# #############################################################################
from __future__ import absolute_import


# -----------------------------------------------------------------------------
#   Imports
# -----------------------------------------------------------------------------
import logging
import threading


# -----------------------------------------------------------------------------
#   Globals
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


class CLI(object):
    """ simplistic CLI (ipython based or fallback), that can be run in a thread
        or in the main thread.
    """
    input_func = input

    def __init__(self, options=None, namespace=None, quit_func=None):
        self._cli_thread = None
        self._options = options

        # create a name space if not passed to __init__ and ensure, that
        # a lock (named __lock) is added to the name space if not already provided
        if namespace is None:
            namespace = {}
        self.namespace = namespace

        # a function to be called when the CLI will be quit.
        # it can do some cleanup work if required
        self._quit_function = quit_func

    def set_quit_function(self, func):
        self._quit_function = func

    def run(self):
        """ allows to run an ipython shell with the CLI's context vars """
        namespace = self.namespace
        try:
            from IPython.terminal.embed import InteractiveShellEmbed
            use_ipython = True
            logger.debug("CLI using ipython")
        except ImportError:
            use_ipython = False
            logger.debug("CLI using basic fallback")

        if use_ipython:
            shell = InteractiveShellEmbed(user_ns=namespace)
            shell()

        else:
            self.mini_shell(namespace=namespace)

        if self._quit_function:
            self._quit_function(self)

    def mini_shell(self, namespace):
        """ Rather lousy Python shell for debugging.
            Just in case ipython is not installed or has the wrong version
        """
        while True:
            cmd_line = self.input_func('-->')
            upper_stripped = cmd_line.strip().upper()
            shall_quit = (upper_stripped == 'Q' or upper_stripped == 'QUIT')
            if shall_quit:
                break
            try:
                eval(compile(cmd_line, '<string>', 'single'), namespace)  # pylint: disable=W0122,C0301
            except Exception as exc:  # pylint: disable=W0703
                logger.error('ERROR: %r' % exc)

        print("END OF CLI")

    def run_as_thread(self, name='cli', daemon=True):
        """ start CLI as a thread.
            This is needed for some programs where other code MUST be started in the main thread
            (e.g. PyQT applications)
        """
        self._cli_thread = cli_thread = threading.Thread(target=self.run, name=name)
        cli_thread.daemon = daemon
        cli_thread.start()



"""  Functionality, that can be used bys eveal test cases

"""

import asyncio
import datetime
import logging
import os
import threading
import time

from importlib import reload

from pypeman import channels
from pypeman import endpoints
from pypeman import nodes
from pypeman import message

logger = logging.getLogger(__name__)


def setup_settings(module):
    """ helper allows to have specific settings for a test
    """
    os.environ['PYPEMAN_SETTINGS_MODULE'] = module
    import pypeman.default_settings
    import pypeman.conf
    reload(pypeman.default_settings)
    reload(pypeman.conf)
    import pypeman.tests.test_app.settings as tst_settings
    reload(tst_settings)
    from pypeman.conf import settings  # noqa: F401


def teardown_settings():
    """ helper allowing to reset settings to default
    """
    os.environ['PYPEMAN_SETTINGS_MODULE'] = 'pypeman.tests.settings.test_settings_default'

    # TODO: try later with del sys.modules[module_name]
    import logging
    logger = logging.getLogger()  # noqa: F841
    import pypeman.default_settings
    reload(pypeman.default_settings)
    try:
        import pypeman.conf
        reload(pypeman.conf)  # ########
    except ImportError:
        pass


default_message_content = """{"test":1}"""
default_message_meta = {'question': 'unknown'}


def generate_msg(timestamp=None, message_content=default_message_content,
                 message_meta=None, with_context=False):
    """ generates a default message """
    m = message.Message()
    if timestamp:
        if isinstance(timestamp, tuple):
            m.timestamp = datetime.datetime(*timestamp)
        else:  # assume it's a datetime object
            m.timestamp = timestamp
    else:  # just use current time
        m.timestamp = datetime.datetime.utcnow()

    m.payload = message_content

    if message_meta is None:
        m.meta = dict(default_message_meta)
    else:
        m.meta = message_meta

    if with_context:
        # Add context message
        mctx = generate_msg(message_content={'question': 'known'}, message_meta={'answer': 43})
        m.add_context('test', mctx)

    return m


class TstException(Exception):
    """ custom Exception """


class LongNode(nodes.ThreadNode):
    def process(self, msg):
        time.sleep(1)
        return msg


class StoreNode(nodes.BaseNode):
    """
    just store each payload in a buffer
    uses a threading lock to allow tests from another thread
    """
    def __init__(self, *args, **kwargs):
        super().__init__(args, **kwargs)
        self.lock = kwargs.pop("lock", None) or threading.Lock()
        self.payloads = []

    def process(self, msg):
        with self.lock:
            self.payloads.append(msg.payload)
        return msg


class SimpleTestNode(nodes.BaseNode):
    """ simple node, that can be used for unit testing
    """
    def __init__(self, *args, logger=None, loop=None, **kwargs):
        """
            :param delay: delay in milliseconds to simulate processing time
            :param logger: allows to pass a custom logger for tracing
        """
        self.delay = kwargs.pop('delay', 0)
        self.async_delay = kwargs.pop('async_delay', None)
        self.logger = logger or logging.getLogger(__name__)
        self.loop = loop
        super().__init__(*args, **kwargs)

        # Used to test if node is processed during test
        self.processed = False

    async def process(self, msg):
        if self.delay:
            time.sleep(self.delay)
        if self.async_delay is not None:
            await asyncio.sleep(self.async_delay)
        self.logger.info("Process done: %s", msg)
        self.processed = True
        return msg


class TstNode(nodes.BaseNode):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Used to test if node is processed during test

    def process(self, msg):
        print("Process %s" % self.name)
        return msg


class ExceptNode(TstNode):
    # This node raises an exception
    def process(self, msg):
        super().process(msg)
        raise TstException()


class MllPChannelTestThread(threading.Thread):
    """
    Start an MLLP Channel in a thread.
    Useful for test purpose
    can be stopped by calling .kill() and .join() methods

    MllpChannel Thread will auto destruct it to avoid infinite loops
    You can disable this feature with setting timeout arg to 0 or None

    params:
        chan_name: (str) name of the mllpChannel
        host: (str) default to '0.0.0.0'
        port: (int) default to 21000
        timeout: (float) default to 5 (5s)

    """
    def __init__(self, chan_name, host="0.0.0.0", port=21000, timeout=5):
        self.timeout = timeout
        self.async_loop = asyncio.new_event_loop()
        sock = f"{host}:{port}"
        endpoint = endpoints.MLLPEndpoint(loop=self.async_loop, sock=sock)
        self.chan = channels.MLLPChannel(
            name=chan_name, endpoint=endpoint, loop=self.async_loop)
        self.auto_kill_delayedtask = None if not timeout else threading.Timer(timeout, self._auto_kill)
        super().__init__(target=self.async_loop.run_forever)

    def start(self, *args, **kwargs):
        self.auto_kill_delayedtask.start()
        self.async_loop.run_until_complete(self.chan.start())
        self.async_loop.run_until_complete(self.chan.mllp_endpoint.start())
        super().start(*args, **kwargs)

    def _auto_kill(self):
        """
        Used to auto kill mllpChannel if it tooks too long time (call self.kill)
        """
        logger.warning(
            "mllp chan %s keeped alive too long time (timeout=%f), killing it",
            self.chan.name, self.timeout)
        self.kill()

    def kill(self):
        logger.debug("killing MllpChanTestThread %s", self.chan.name)
        if self.auto_kill_delayedtask and self.auto_kill_delayedtask.is_alive():
            self.auto_kill_delayedtask.cancel()
        pending_tasks = asyncio.all_tasks(loop=self.async_loop)
        for pending in pending_tasks:
            pending.cancel()
        self.async_loop.call_soon_threadsafe(self.async_loop.stop)



"""
UnitTests checking, that debugging works as expected
"""

import asyncio
import logging
import os
import time

from pypeman.channels import BaseChannel

from pypeman.test import TearDownProjectTestCase as TestCase
from pypeman.tests.common import SimpleTestNode
from pypeman.tests.common import generate_msg
from pypeman.tests.common import setup_settings
from pypeman.tests.common import teardown_settings


SETTINGS_MODULE = 'pypeman.tests.test_app.settings'
SETTINGS_MODULE2 = 'pypeman.tests.test_app.settings2'


class EvtLoopMixin:
    def setup_evt_loop(self, slow_cb_duration=0.1):
        # Create class event loop used for tests to avoid failing
        # previous tests to impact next test ? (Not shure)
        self.loop = loop = asyncio.new_event_loop()
        loop.slow_callback_duration = slow_cb_duration
        # Remove thread event loop to be sure we are not using
        # another event loop somewhere
        asyncio.set_event_loop(None)


class LoggingTests(TestCase, EvtLoopMixin):
    def setUp(self):
        setup_settings(SETTINGS_MODULE)  # adapt config
        from pypeman.conf import settings
        settings.LOGGING  # force loading of cfg
        # create logger and add custom handler
        self.logger = logging.getLogger()
        print("got logger")
        from pypeman.helpers.logging import DebugLogHandler
        self.loghandler = handler = DebugLogHandler()
        self.logger.handlers.append(handler)
        print("added handler")

    def tearDown(self):
        super().tearDown()
        setup_settings(SETTINGS_MODULE)
        teardown_settings()

    def test_log(self):
        """ can modify log config
            changing log config may be important for other tests.
            thus let's try to change it and see its impact.
        """

        self.setup_evt_loop()
        logger = self.logger  # get default test logger
        handler = self.loghandler  # get test log handler

        cnt0 = handler.num_entries()
        # test second logger
        logger.debug("DEBUG")
        self.assertEqual(handler.num_entries()-cnt0, 0, "A0: expected 0 log entry")
        logger.info("INFO")
        self.assertEqual(handler.num_entries()-cnt0, 0, "A1: expected 0 log entry")
        logger.warning("WARNING")
        self.assertEqual(handler.num_entries()-cnt0, 1, "A2: expected 1 log entry")
        logger.error("ERROR")

        # now reconfigure logging
        from pypeman.helpers.logging import DebugLogHandler
        setup_settings(SETTINGS_MODULE2)  # adapt config
        from pypeman.conf import settings
        settings.LOGGING  # force loading of cfg
        # create logger and add custom handler
        logger = logging.getLogger()
        handler = DebugLogHandler()
        logger.handlers.append(handler)

        cnt0 = handler.num_entries()
        # test first logger
        logger.debug("DEBUG")
        self.assertEqual(handler.num_entries()-cnt0, 1, "B: expected 1 log entry")
        logger.info("INFO")
        self.assertEqual(handler.num_entries()-cnt0, 2, "B: expected 2 log entries")

    def test_no_log(self):
        pass


class MainLoopTests(TestCase, EvtLoopMixin):
    def setUp(self):
        """ """
        self.prev_debug_flag = os.environ.get('PYTHONASYNCIODEBUG', None)
        setup_settings(SETTINGS_MODULE)  # adapt config
        # from pypeman.conf import settings
        # _logging = settings.LOGGING  # force loading of cfg

        # create logger and add custom handler
        # self.logger = logging.getLogger()
        # print("got logger")
        # from pypeman.helpers.logging import DebugLogHandler
        # self.loghandler = handler = DebugLogHandler()
        # self.logger.handlers.append(handler)
        # print("added handler")

    def tearDown(self):
        super().tearDown()
        if self.prev_debug_flag is not None:
            os.environ['PYTHONASYNCIODEBUG'] = self.prev_debug_flag
        else:
            del os.environ['PYTHONASYNCIODEBUG']
        setup_settings(SETTINGS_MODULE)
        teardown_settings()

    def test_loop_slow(self):
        """ main loop logs slow tasks """
        import pypeman.debug
        pypeman.debug.enable_slow_log_stats()
        stats = pypeman.debug.stats
        self.setup_evt_loop()

        # logger = self.logger # get default test logger
        # handler = self.loghandler # get test log handler

        tst_logger = logging.getLogger('tests.debug.main_loop.slow')
        print("HANDLERS", tst_logger.handlers)

        loop = self.loop
        chan = BaseChannel(name="test_loop_slow", loop=loop)
        n1 = SimpleTestNode(delay=0.01, async_delay=0, logger=tst_logger, loop=loop)
        n2 = SimpleTestNode(delay=0.12, async_delay=0, logger=tst_logger, loop=loop)
        n3 = SimpleTestNode(delay=0.11, async_delay=0, logger=tst_logger, loop=loop)
        chan.add(n1)
        chan.add(n2)
        chan.add(n3)

        msg = generate_msg()
        # Launch channel processing
        t0 = time.time()
        print("start channel %.4f" % (time.time() - t0))

        self.loop.run_until_complete(chan.start())
        print("channel started %.4f" % (time.time() - t0))

        self.loop.run_until_complete(chan.handle(msg))
        print("msg handled %.4f" % (time.time() - t0))
        # handler.show_entries()

        self.assertEqual(len(stats), 2, "should have 2 slow tasks, not %d" % len(stats))
        self.loop.close()

    def test_loop_slow2(self):
        """ slow tasks delay can be configured """
        import pypeman.debug
        pypeman.debug.enable_slow_log_stats()
        self.setup_evt_loop(slow_cb_duration=0.05)
        stats = pypeman.debug.stats

        # logger = self.logger # get default test logger
        # handler = self.loghandler # get test log handler

        tst_logger = logging.getLogger('tests.debug.main_loop.slow')
        print(tst_logger.handlers)

        loop = self.loop
        chan = BaseChannel(name="test_loop_slow2", loop=loop)
        n1 = SimpleTestNode(delay=0.03, async_delay=0, logger=tst_logger, loop=loop)
        n2 = SimpleTestNode(delay=0.06, logger=tst_logger, loop=loop)
        chan.add(n1)
        chan.add(n2)

        msg = generate_msg()
        # Launch channel processing
        self.loop.run_until_complete(chan.start())
        self.loop.run_until_complete(chan.handle(msg))
        # handler.show_entries()
        self.assertEqual(len(stats), 1, "should have 1 slow tasks, not %d" % len(stats))
        pypeman.debug.show_slow_log_stats()
        self.loop.close()
